from typing import Dict, List

import asf_search as asf
from asf_search import ASFSearchResults
import defusedxml.ElementTree as DefusedETree
import xml.etree.ElementTree as ETree
import json
import shapely.wkt as WKT
import requests
import csv

from shapely.geometry import Polygon
from shapely.wkt import loads
from shapely.geometry import shape
from shapely.geometry.base import BaseGeometry
from asf_search.CMR.translate import try_parse_date
from asf_search.constants import PLATFORM
from asf_search import ASF_LOGGER
import re

from asf_search.exceptions import ASFSearchError

# when this replaces SearchAPI change values to cached
API_URL = 'https://api.daac.asf.alaska.edu/services/search/param?'


def run_test_output_format(results: ASFSearchResults):
    # search results are always sorted this way when returned from asf_search.search(),
    # but not all test case resources are
    results.sort(key=lambda p: (p.properties['stopTime'], p.properties['fileID']), reverse=True)
    product_list_str = ','.join([product.properties['fileID'] for product in results])

    results.searchComplete = True

    for output_type in ['csv', 'kml', 'metalink', 'jsonlite', 'jsonlite2', 'geojson']:
        expected = get_SearchAPI_Output(product_list_str, output_type)
        if output_type == 'csv':
            check_csv(results, expected)
        if output_type == 'kml':
            check_kml(results, expected)
        elif output_type == 'metalink':
            check_metalink(results, expected)
        elif output_type in ['jsonlite', 'jsonlite2']:
            check_jsonLite(results, expected, output_type)
        elif output_type == 'geojson':
            check_geojson(results)


def check_metalink(results: ASFSearchResults, expected_str: str):
    actual = ''.join([line for line in results.metalink()])

    actual_tree = DefusedETree.fromstring(actual)
    expected_tree = DefusedETree.fromstring(expected_str)

    canon_actual = ETree.canonicalize(DefusedETree.tostring(actual_tree), strip_text=True)
    canon_expected = ETree.canonicalize(DefusedETree.tostring(expected_tree), strip_text=True)

    assert canon_actual == canon_expected


def check_kml(results: ASFSearchResults, expected_str: str):
    namespaces = {'kml': 'http://www.opengis.net/kml/2.2'}
    placemarks_path = './/kml:Placemark'
    expected_root = DefusedETree.fromstring(expected_str)
    expected_placemarks = expected_root.findall(placemarks_path, namespaces)

    actual_root = DefusedETree.fromstring(''.join([block for block in results.kml()]))
    actual_placemarks = actual_root.findall(placemarks_path, namespaces)

    # Check polygons for equivalence (asf-search starts from a different pivot)
    # and remove them from the kml so we can easily compare the rest of the placemark data
    for expected_placemark, actual_placemark in zip(expected_placemarks, actual_placemarks):
        expected_polygon = expected_placemark.findall('./*')[-1]
        actual_polygon = actual_placemark.findall('./*')[-1]

        expected_coords = get_coordinates_from_kml(DefusedETree.tostring(expected_polygon))
        actual_coords = get_coordinates_from_kml(DefusedETree.tostring(actual_polygon))

        assert Polygon(expected_coords).equals(Polygon(actual_coords))

        expected_placemark.remove(expected_polygon)
        actual_placemark.remove(actual_polygon)

    # Get canonicalize xml strings so minor differences are normalized
    actual_canon = ETree.canonicalize(DefusedETree.tostring(actual_root), strip_text=True)
    expected_canon = ETree.canonicalize(DefusedETree.tostring(expected_root), strip_text=True)

    date_pattern = r'\>(?P<variable>[\w ]*time|Time): *(?P<datestring>[^\<]*)\<'

    actual_dates = re.findall(date_pattern, actual_canon, re.MULTILINE)
    expected_date = re.findall(date_pattern, expected_canon, re.MULTILINE)

    for idx, match in enumerate(actual_dates):
        date_str, date_value = match
        assert expected_date[idx][0] == date_str
        assert try_parse_date(expected_date[idx][1]) == try_parse_date(date_value)

    actual_canon = re.sub(date_pattern, '', actual_canon)
    expected_canon = re.sub(date_pattern, '', expected_canon)
    assert actual_canon == expected_canon


def get_coordinates_from_kml(data: str):
    namespaces = {'kml': 'http://www.opengis.net/kml/2.2'}

    coords = []
    coords_lon_lat_path = './/kml:outerBoundaryIs/kml:LinearRing/kml:coordinates'
    root = DefusedETree.fromstring(data)

    coordinates_elements = root.findall(coords_lon_lat_path, namespaces)
    for lon_lat_z in coordinates_elements[0].text.split('\n'):
        if len(lon_lat_z.split(',')) == 3:
            lon, lat, _ = lon_lat_z.strip().split(',')
            coords.append([float(lon), float(lat)])

    return coords


def check_csv(results: ASFSearchResults, expected_str: str):
    expected = [product for product in csv.reader(expected_str.split('\n')) if product != []]
    # actual = [prod for prod in csv.reader(''.join([s for s in results.csv()]).split('\n')) if prod != []]

    expected = csv.DictReader(expected_str.split('\n'))
    actual = csv.DictReader([s for s in results.csv()])

    for actual_row, expected_row in zip(actual, expected):
        actual_dict = dict(actual_row)
        expected_dict = dict(expected_row)

        for key in expected_dict.keys():
            if expected_dict[key] in ['None', None, '']:
                assert actual_dict[key] in ['None', None, '']
            else:
                try:
                    expected_value = float(expected_dict[key])
                    actual_value = float(actual_dict[key])
                    assert (
                        expected_value == actual_value
                    ), f"expected '{expected_dict[key]}' for key '{key}', got '{actual_dict[key]}'"
                except ValueError:
                    try:
                        expected_date = try_parse_date(expected_dict[key])
                        actual_date = try_parse_date(actual_dict[key])
                        assert (
                            expected_date == actual_date
                        ), f"Expected date '{expected_date}' for key '{key}', got '{actual_date}'"
                    except ValueError:
                        assert (
                            expected_dict[key] == actual_dict[key]
                        ), f"expected '{expected_dict[key]}' for key '{key}', got '{actual_dict[key]}'"


def check_jsonLite(results: ASFSearchResults, expected_str: str, output_type: str):
    jsonlite2 = output_type == 'jsonlite2'

    expected = json.loads(expected_str)['results']

    if jsonlite2:
        wkt_key = 'w'
        wkt_unwrapped_key = 'wu'
        start_time_key = 'st'
        stop_time_key = 'stp'
    else:
        wkt_key = 'wkt'
        wkt_unwrapped_key = 'wkt_unwrapped'
        start_time_key = 'startTime'
        stop_time_key = 'stopTime'

    actual = json.loads(''.join(results.jsonlite2() if jsonlite2 else results.jsonlite()))[
        'results'
    ]

    for idx, expected_product in enumerate(expected):
        wkt = expected_product.pop(wkt_key)
        wkt_unwrapped = expected_product.pop(wkt_unwrapped_key)

        startTime = expected_product.pop(start_time_key)
        stopTime = expected_product.pop(stop_time_key)

        for key in expected_product.keys():
            assert actual[idx][key] == expected_product[key]

        assert WKT.loads(actual[idx][wkt_key]).equals(WKT.loads(wkt))
        assert WKT.loads(actual[idx][wkt_unwrapped_key]).equals(WKT.loads(wkt_unwrapped))
        assert actual[idx][start_time_key] == try_parse_date(startTime)
        assert actual[idx][stop_time_key] == try_parse_date(stopTime)


def check_geojson(results: ASFSearchResults):
    expected = results.geojson()
    actual = asf.export.results_to_geojson(results)

    assert json.loads(''.join(actual)) == expected


def get_SearchAPI_Output(product_list: List[str], output_type: str) -> List[Dict]:
    response = requests.get(API_URL, [('product_list', product_list), ('output', output_type)])
    response.raise_for_status()

    expected = response.text

    return expected


def run_test_ASFSearchResults_intersection(wkt: str):
    wrapped, unwrapped, _ = asf.validate_wkt(wkt)
    unchanged_aoi = loads(wkt)  # sometimes geometries don't come back with wrapping in mind

    # exclude SMAP products
    platforms = [PLATFORM.SENTINEL1, PLATFORM.UAVSAR]

    def overlap_check(s1: BaseGeometry, s2: BaseGeometry):
        return s1.overlaps(s2) or s1.touches(s2) or s2.distance(s1) <= 0.005 or s1.dwithin

    for platform in platforms:
        asf.constants.INTERNAL.CMR_TIMEOUT = 120
        try:
            results = asf.geo_search(intersectsWith=wkt, platform=platform, maxResults=250)
        except ASFSearchError as exc:
            asf.constants.INTERNAL.CMR_TIMEOUT = 30
            if str(exc).startswith("Connection Error (Timeout):"):
                ASF_LOGGER.warning('CMR timeout while running intersection test')
                continue
            else:
                raise BaseException(
                    f'Failed to perform intersection test with wkt: {wkt}\nplatform: {platform}.\nOriginal exception: {exc}'
                )

        asf.constants.INTERNAL.CMR_TIMEOUT = 30
        for product in results:
            if shape(product.geometry).is_valid:
                product_geom_wrapped, product_geom_unwrapped, _ = asf.validate_wkt(
                    shape(product.geometry)
                )
                original_shape = unchanged_aoi

                assert (
                    overlap_check(product_geom_wrapped, wrapped)
                    # or overlap_check(product_geom_wrapped, original_shape)
                    or overlap_check(product_geom_unwrapped, original_shape)
                ), f"OVERLAP FAIL: {product.properties['sceneName']}, {product.geometry} \nproduct: {product_geom_wrapped.wkt} \naoi: {wrapped.wkt}"



import inspect
import os
from types import GeneratorType
from urllib import parse
import xml.etree.ElementTree as ETree

from asf_search import ASF_LOGGER
from asf_search.export.export_translators import ASFSearchResults_to_properties_list


def results_to_metalink(results):
    ASF_LOGGER.info('Started translating results to metalink format')

    if inspect.isgeneratorfunction(results) or isinstance(results, GeneratorType):
        return MetalinkStreamArray(results)

    return MetalinkStreamArray([results])


class MetalinkStreamArray(list):
    def __init__(self, results):
        self.pages = results
        self.len = 1
        self.header = (
            '<?xml version="1.0"?>'
            '<metalink xmlns="http://www.metalinker.org/" version="3.0">\n'
            '<publisher><name>Alaska Satellite Facility</name><url>http://www.asf.alaska.edu/</url></publisher>\n'  # noqa F401
            '<files>'
        )

        self.footer = '\n</files>\n</metalink>'

    def get_additional_fields(self, product):
        return {}

    def __iter__(self):
        return self.streamPages()

    def __len__(self):
        return self.len

    def streamPages(self):
        yield self.header

        completed = False
        for page_idx, page in enumerate(self.pages):
            ASF_LOGGER.info(f'Streaming {len(page)} products from page {page_idx}')
            completed = page.searchComplete

            properties_list = ASFSearchResults_to_properties_list(page, self.get_additional_fields)
            yield from [self.getItem(p) for p in properties_list]

        if not completed:
            ASF_LOGGER.warn('Failed to download all results from CMR')

        yield self.footer

        ASF_LOGGER.info(f'Finished streaming {self.getOutputType()} results')

    def getOutputType(self) -> str:
        return 'metalink'

    def getItem(self, p):
        file = ETree.Element('file', attrib={'name': p['fileName']})
        resources = ETree.Element('resources')

        url = ETree.Element('url', attrib={'type': 'http'})
        url.text = p['url']
        resources.append(url)
        file.append(resources)

        if p.get('md5sum') and p.get('md5sum') != 'NA':
            verification = ETree.Element('verification')
            if isinstance(p.get('md5sum'), dict):
                a = parse.urlparse(p['url'])
                file_name = os.path.basename(a.path)
                md5_entry = p['md5sum'].get(file_name)
                h = ETree.Element('hash', {'type': 'md5'})
                if md5_entry is not None:
                    h.text=md5_entry
                    verification.append(h)
            else:
                h = ETree.Element('hash', {'type': 'md5'})
                h.text = p['md5sum']
                verification.append(h)
            file.append(verification)

        if p['bytes'] and p['bytes'] != 'NA':
            size = ETree.Element('size')
            if isinstance(p.get('bytes'), dict):
                a = parse.urlparse(p['url'])
                file_name = os.path.basename(a.path)
                bytes_entry = p['bytes'].get(file_name)
                if bytes_entry is not None:
                    size.text=str(bytes_entry['bytes'])
                else:
                    size.text = str(p['bytes'])
            else:
                size.text = str(p['bytes'])
            file.append(size)

        return '\n' + (8 * ' ') + ETree.tostring(file, encoding='unicode')

    def indent(self, elem, level=0):
        # Only Python 3.9+ has a built-in indent function for element tree.
        # https://stackoverflow.com/a/33956544
        i = '\n' + level * '  '
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + '  '
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for elem in elem:
                self.indent(elem, level + 1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i



from typing import Dict, List
from asf_search import ASFSearchOptions, ASFSession, FileDownloadType, search

from asf_search.exceptions import ASFAuthenticationError

from ASFProduct.test_ASFProduct import (
    run_test_ASFProduct,
    run_test_ASFProduct_download,
    run_test_product_get_stack_options,
    run_test_stack,
)
from ASFSearchOptions.test_ASFSearchOptions import run_test_ASFSearchOptions
from ASFSearchResults.test_ASFSearchResults import (
    run_test_ASFSearchResults_intersection,
)
from ASFSession.test_ASFSession import (
    run_auth_with_cookiejar,
    run_auth_with_creds,
    run_auth_with_token,
    run_test_asf_session_rebuild_auth,
)
from BaselineSearch.test_baseline_search import (
    run_test_stack_from_id,
    run_test_stack_from_product,
    run_test_calc_temporal_baselines,
    run_get_stack_opts_invalid_insarStackId,
    run_test_get_unprocessed_stack_params,
    run_test_get_preprocessed_stack_params,
)
from Search.test_search import (
    run_test_ASFSearchResults,
    run_test_build_subqueries,
    run_test_dataset_search,
    run_test_keyword_aliasing_results,
    run_test_search,
    run_test_search_http_error,
)
from Search.test_search_generator import (
    run_test_search_generator,
    run_test_search_generator_multi,
)
from CMR.test_MissionList import run_test_get_project_names


from pytest import raises
from unittest.mock import patch

import os
import pathlib
import yaml

from WKT.test_validate_wkt import (
    run_test_search_wkt_prep,
    run_test_validate_wkt_get_shape_coords,
    run_test_validate_wkt_clamp_geometry,
    run_test_validate_wkt_valid_wkt,
    run_test_validate_wkt_convex_hull,
    run_test_validate_wkt_counter_clockwise_reorientation,
    run_test_validate_wkt_invalid_wkt_error,
    run_test_validate_wkt_merge_overlapping_geometry,
    run_test_simplify_aoi,
)
from ASFSearchOptions.test_ASFSearchOptions import (
    run_test_ASFSearchOptions_validator,
    run_test_validator_map_validate,
)
from BaselineSearch.Stack.test_stack import (
    run_test_find_new_reference,
    run_test_get_baseline_from_stack,
    run_test_get_default_product_type,
    run_test_valid_state_vectors,
)
from asf_search.search.search_generator import as_ASFProduct

from download.test_download import (
    run_test_download_url,
    run_test_download_url_auth_error,
)
from Serialization.test_serialization import run_test_serialization
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor

from ASFProduct.test_ASFSubproduct import run_test_ASFSubproduct
from tests.Search.test_collection_attributes import run_test_collection_attributes


# asf_search.ASFProduct Tests
def test_ASFProduct(**args) -> None:
    """
    Tests Basic ASFProduct with mock searchAPI response
    """
    test_info = args['test_info']
    geographic_response = get_resource(test_info['products'])
    run_test_ASFProduct(geographic_response)

def test_ASFSubproduct(**args) -> None:
    """
    Tests ASFProduct subclasses for properties and basic functionality
    """
    session = args["config"].getoption("authenticated_session")

    test_info = args['test_info']
    scene_names = test_info['scenes']
    expected_subclass = test_info['expected_subclass']
    opts = ASFSearchOptions(**test_info.get('opts', {}), session=session)

    run_test_ASFSubproduct(scene_names=scene_names, expected_subclass=expected_subclass, opts=opts)
    
def test_ASFProduct_Stack(**args) -> None:
    """
    Tests ASFProduct.stack() with reference and corresponding stack
    Checks for temporalBaseline order,
    asserting the stack is ordered by the scene's temporalBaseline (in ascending order)
    """
    test_info = args['test_info']
    reference = get_resource(test_info['product'])
    preprocessed_stack = get_resource(test_info['preprocessed_stack'])
    processed_stack = get_resource(test_info['processed_stack'])
    run_test_stack(reference, preprocessed_stack, processed_stack)


def test_ASFProduct_get_stack_options(**args) -> None:
    test_info = args['test_info']
    reference = get_resource(test_info['product'])
    options = get_resource(test_info['options'])

    run_test_product_get_stack_options(reference, options)


def test_ASFProduct_download(**args) -> None:
    test_info = args['test_info']
    reference = get_resource(test_info['product'])
    filename = test_info['filename']
    filetype_raw = test_info['filetype']
    additional_urls = test_info['additionalUrls']

    if filetype_raw == 1:
        filetype = FileDownloadType.DEFAULT_FILE
    elif filetype_raw == 2:
        filetype = FileDownloadType.ADDITIONAL_FILES
    else:
        filetype = FileDownloadType.ALL_FILES

    run_test_ASFProduct_download(reference, filename, filetype, additional_urls)


# asf_search.ASFSession Tests
def test_ASFSession_Error(**args) -> None:
    """
    Test ASFSession.auth_with_creds for sign in errors
    """
    test_info = args['test_info']
    username = test_info['username']
    password = test_info['password']
    with patch('asf_search.ASFSession.get') as mock_get:
        mock_get.return_value = 'Error'

        with raises(ASFAuthenticationError):
            run_auth_with_creds(username, password)


def test_ASFSession_Token_Error(**args) -> None:
    """
    Test ASFSession.auth_with_token for sign in errors
    """
    test_info = args['test_info']
    token = test_info['token']

    with raises(ASFAuthenticationError):
        run_auth_with_token(token)


def test_ASFSession_Cookie_Error(**args) -> None:
    """
    Test ASFSession.auth_with_cookie for sign in errors
    """
    test_info = args['test_info']
    cookies = test_info['cookies']

    with raises(ASFAuthenticationError):
        run_auth_with_cookiejar(cookies)


def test_asf_session_rebuild_auth(**args) -> None:
    """
    Test asf_search.ASFSession.rebuild_auth
    When redirecting from an ASF domain, only accept
    domains listed in ASFSession.AUTH_DOMAINS
    """
    test_info = args['test_info']
    original_domain = test_info['original_domain']
    response_domain = test_info['response_domain']
    response_code = test_info['response_code']
    final_token = test_info['final_token']

    run_test_asf_session_rebuild_auth(original_domain, response_domain, response_code, final_token)


# asf_search.search.baseline_search Tests
def test_get_preprocessed_stack_params(**args) -> None:
    """
    Test asf_search.search.baseline_search.get_stack_opts with a reference scene
    that's part of a pre-calculated platform, asserting that get_stack_opts returns an object with two parameters
    \n1. processingLevel
    \n2. insarStackId
    """
    test_info = args['test_info']
    reference = get_resource(test_info['product'])

    run_test_get_preprocessed_stack_params(reference)


def test_get_unprocessed_stack_params(**args) -> None:
    """
    Test asf_search.search.baseline_search.get_stack_opts with a reference scene
    that's not part of a pre-calculated platform, asserting that get_stack_opts returns an object with seven parameters
    """
    test_info = args['test_info']
    reference = get_resource(test_info['product'])

    run_test_get_unprocessed_stack_params(reference)


def test_get_stack_opts_invalid_insarStackId(**args) -> None:
    """
    Test asf_search.search.baseline_search.get_stack_opts with a the reference scene's
    insarStackID set to an invalid value, and asserting an ASFBaselineError is raised
    """
    test_info = args['test_info']
    reference = get_resource(test_info['product'])

    run_get_stack_opts_invalid_insarStackId(reference)


def test_temporal_baseline(**args) -> None:
    """
    Test asf_search.search.baseline_search.calc_temporal_baselines, asserting mutated baseline stack
    is still the same length and that each product's properties contain a temporalBaseline key
    """
    test_info = args['test_info']
    reference = get_resource(test_info['product'])
    stack = get_resource(test_info['stack'])
    run_test_calc_temporal_baselines(reference, stack)


def test_stack_from_product(**args) -> None:
    """
    Test asf_search.search.baseline_search.stack_from_product, asserting stack returned is ordered
    by temporalBaseline value in ascending order
    """
    test_info = args['test_info']
    reference = get_resource(test_info['product'])
    stack = get_resource(test_info['stack'])

    run_test_stack_from_product(reference, stack)


def test_stack_from_id(**args) -> None:
    """
    Test asf_search.search.baseline_search.stack_from_id, asserting stack returned is ordered
    by temporalBaseline value in ascending order
    """
    test_info = args['test_info']
    stack_id = test_info['stack_id']
    stack_reference_data = test_info['stack_reference']
    stack_data = test_info['stack']
    dataset = test_info.get('dataset', None)

    stack_reference = get_resource(stack_reference_data)
    stack = []

    if stack_data != []:
        stack = get_resource(stack_data)

    opts = ASFSearchOptions(dataset=dataset)
    run_test_stack_from_id(stack_id, stack_reference, stack, opts=opts)


# asf_search.ASFSearchResults Tests
def test_ASFSearchResults(**args) -> None:
    """
    Test asf_search.ASFSearchResults, asserting initialized values,
    and geojson response returns object with type FeatureCollection
    """
    test_info = args['test_info']
    search_response = get_resource(test_info['response'])

    run_test_ASFSearchResults(search_response)


# asf_search.search Tests
def test_ASFSearch_Search(**args) -> None:
    """
    Test asf_search.search, asserting returned value is expected result
    """
    test_info = args['test_info']
    parameters = get_resource(test_info['parameters'])
    answer = get_resource(test_info['answer'])

    run_test_search(parameters, answer)


def test_ASFSearch_Search_Generator(**args) -> None:
    test_info = args['test_info']
    params = get_resource(test_info['parameters'])

    if isinstance(params, list):
        opts = []
        for p in params:
            opts.append(ASFSearchOptions(**p))

        run_test_search_generator_multi(opts)

    else:
        run_test_search_generator(ASFSearchOptions(**params))


def test_ASFSearch_Search_Error(**args) -> None:
    """
    Test asf_search.search errors,
    asserting server and client errors are raised
    """
    test_info = args['test_info']
    parameters = test_info['parameters']
    report = test_info['report']
    error_code = test_info['status_code']

    run_test_search_http_error(parameters, error_code, report)


def test_wkt_validation_Invalid_WKT_Error(**args) -> None:
    """
    Test asf_search.wkt errors,
    asserting wkt validation errors are raised
    """
    test_info = args['test_info']
    wkt = get_resource(test_info['wkt'])
    run_test_validate_wkt_invalid_wkt_error(wkt)


def test_wkt_validation_WKT_Valid(**args) -> None:
    """
    Test asf_search.validate_wkt, asserting expected wkts are returned
    """
    test_info = args['test_info']
    wkt = get_resource(test_info['wkt'])
    validated_wkt = get_resource(test_info['validated-wkt'])
    run_test_validate_wkt_valid_wkt(wkt, validated_wkt)


def test_wkt_validation_WKT_clamp_geometry(**args) -> None:
    """
    Test asf_search.validate_wkt._get_clamped_and_wrapped_geometry, asserting the amount of clamped and wrapped coordinates
    """
    test_info = args['test_info']
    wkt = get_resource(test_info['wkt'])
    clamped_wkt = get_resource(test_info['clamped-wkt'])
    clamped_count = get_resource(test_info['clamped-count'])
    wrapped_count = get_resource(test_info['wrapped-count'])
    run_test_validate_wkt_clamp_geometry(wkt, clamped_wkt, clamped_count, wrapped_count)


def test_wkt_validation_convex_hull(**args) -> None:
    """
    Test asf_search.validate_wkt._get_convex_hull, asserting convex hulls producted are expected
    """
    test_info = args['test_info']
    wkt = get_resource(test_info['wkt'])
    convex_wkt = get_resource(test_info['convex-wkt'])
    run_test_validate_wkt_convex_hull(wkt, convex_wkt)


def test_wkt_validation_merge_overlapping_geometry(**args) -> None:
    """
    Test asf_search.validate_wkt._merge_overlapping_geometry, asserting expected shapes are merged
    """
    test_info = args['test_info']
    wkt = get_resource(test_info['wkt'])
    merged_wkt = get_resource(test_info['merged-wkt'])
    run_test_validate_wkt_merge_overlapping_geometry(wkt, merged_wkt)


def test_wkt_validation_counter_clockwise_reorientation(**args) -> None:
    """
    Test asf_search.validate_wkt._counter_clockwise_reorientation reverses polygon orientation if polygon is wound clockwise,
    and maintains counter-clockwise winding when polygon orientation is correct
    """
    test_info = args['test_info']
    wkt = get_resource(test_info['wkt'])
    cc_wkt = get_resource(test_info['cc-wkt'])
    run_test_validate_wkt_counter_clockwise_reorientation(wkt, cc_wkt)


def test_validate_wkt_get_shape_coords(**args) -> None:
    """
    Test asf_search.validate_wkt._get_shape_coords asserting all coordinates are returned and expected
    """
    test_info = args['test_info']
    wkt = get_resource(test_info['wkt'])
    coords = get_resource(test_info['coordinates'])
    run_test_validate_wkt_get_shape_coords(wkt, coords)


def test_search_wkt_prep(**args) -> None:
    """
    Test asf_search.validate_wkt.wkt_prep, asserting returned shape is correct geometric type and expected shape
    """
    test_info = args['test_info']
    wkt = get_resource(test_info['wkt'])

    run_test_search_wkt_prep(wkt)


def test_simplify_aoi(**args) -> None:
    """
    Test asf_search.validate_wkt.wkt_prep, asserting returned shape is correct geometric type and expected shape
    """
    test_info = args['test_info']
    wkt = get_resource(test_info['wkt'])
    simplified = get_resource(test_info['simplified-wkt'])
    RepairEntries = get_resource(test_info['RepairEntries'])
    run_test_simplify_aoi(wkt, simplified, RepairEntries)


def test_get_platform_campaign_names(**args) -> None:
    test_info = args['test_info']
    cmr_ummjson = get_resource(test_info['cmr_ummjson'])
    campaigns: List[str] = get_resource(test_info['campaigns'])

    run_test_get_project_names(cmr_ummjson, campaigns)


def test_download_url(**args) -> None:
    """
    Test asf_search.download.download_url
    """
    test_info = args['test_info']
    url = test_info['url']
    path = test_info['path']
    filename = test_info['filename']

    if filename == 'error':
        run_test_download_url_auth_error(url, path, filename)
    else:
        run_test_download_url(url, path, filename)


def test_find_new_reference(**args) -> None:
    """
    Test asf_search.baseline.calc.find_new_reference
    """
    test_info = args['test_info']
    stack = get_resource(test_info['stack'])
    output_index = get_resource(test_info['output_index'])

    run_test_find_new_reference(stack, output_index)


def test_get_default_product_type(**args) -> None:
    test_info = args['test_info']
    product = get_resource(test_info['product'])
    product_type = get_resource(test_info['product_type'])

    product = as_ASFProduct({'meta': product['meta'], 'umm': product['umm']}, ASFSession())

    if product.properties.get('sceneName') is None:
        product.properties['sceneName'] = 'BAD_SCENE'

    run_test_get_default_product_type(product, product_type)


def test_get_baseline_from_stack(**args) -> None:
    test_info = args['test_info']
    reference = get_resource(test_info['reference'])
    stack = get_resource(test_info['stack'])
    output_stack = get_resource(test_info['output_stack'])
    error = get_resource(test_info['error'])
    run_test_get_baseline_from_stack(reference, stack, output_stack, error)


def test_valid_state_vectors(**args) -> None:
    test_info = args['test_info']
    reference = get_resource(test_info['reference'])
    output = get_resource(test_info['output'])

    run_test_valid_state_vectors(reference, output)


def test_validator_map_validate(**args) -> None:
    test_info = args['test_info']
    key = get_resource(test_info['key'])
    value = get_resource(test_info['value'])
    output = get_resource(test_info['output'])

    run_test_validator_map_validate(key, value, output)


def test_ASFSearchOptions_validator(**kargs) -> None:
    test_info = kargs['test_info']
    validator_name = get_resource(test_info['validator'])
    param = safe_load_tuple(get_resource(test_info['input']))
    output = safe_load_tuple(get_resource(test_info['output']))
    error = get_resource(test_info['error'])
    run_test_ASFSearchOptions_validator(validator_name, param, output, error)


def test_ASFSearchOptions(**kwargs) -> None:
    run_test_ASFSearchOptions(**kwargs)


def test_ASFSearchResults_get_urls() -> None:
    response = search(
        granule_list=[
            'OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0'
        ]
    )
    actual_urls = response.find_urls()

    expected_urls = [
        'https://cumulus.asf.alaska.edu/s3credentials',
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE.png',
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE.png.md5',
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE_low-res.png',
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE_low-res.png.md5',
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE_thumbnail.png',
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE_thumbnail.png.md5',
        'https://datapool.asf.alaska.edu/RTC/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0.h5',
        'https://datapool.asf.alaska.edu/RTC/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0.h5.md5',
        'https://datapool.asf.alaska.edu/RTC/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0.iso.xml',
        'https://datapool.asf.alaska.edu/RTC/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0.iso.xml.md5',
        'https://datapool.asf.alaska.edu/RTC/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_VH.tif',
        'https://datapool.asf.alaska.edu/RTC/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_VH.tif.md5',
        'https://datapool.asf.alaska.edu/RTC/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_VV.tif',
        'https://datapool.asf.alaska.edu/RTC/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_VV.tif.md5',
        'https://datapool.asf.alaska.edu/RTC/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_mask.tif',
        'https://datapool.asf.alaska.edu/RTC/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_mask.tif.md5',
    ]
    assert actual_urls == expected_urls
    assert response.find_urls('.tif') == [
        'https://datapool.asf.alaska.edu/RTC/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_VH.tif',
        'https://datapool.asf.alaska.edu/RTC/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_VV.tif',
        'https://datapool.asf.alaska.edu/RTC/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_mask.tif',
    ]
    assert response.find_urls(pattern='.*s3credentials') == [
        'https://cumulus.asf.alaska.edu/s3credentials'
    ]
    assert response.find_urls('.tif', directAccess=True) == [
        's3://asf-cumulus-prod-opera-products/OPERA_L2_RTC-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_VH.tif',
        's3://asf-cumulus-prod-opera-products/OPERA_L2_RTC-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_VV.tif',
        's3://asf-cumulus-prod-opera-products/OPERA_L2_RTC-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_mask.tif',
    ]

    assert response.find_urls(pattern=r'.*BROWSE.*') == [
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE.png',
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE.png.md5',
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE_low-res.png',
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE_low-res.png.md5',
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE_thumbnail.png',
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE_thumbnail.png.md5',
    ]
    assert response.find_urls('.png', pattern=r'.*BROWSE.*') == [
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE.png',
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE_low-res.png',
        'https://datapool.asf.alaska.edu/BROWSE/OPERA-S1/OPERA_L2_RTC-S1_T131-279916-IW1_20231202T162856Z_20231202T232622Z_S1A_30_v1.0_BROWSE_thumbnail.png',
    ]


def test_ASFSearchResults_intersection(**kwargs) -> None:
    wkt = get_resource(kwargs['test_info']['wkt'])
    run_test_ASFSearchResults_intersection(wkt)


def test_search_dataset(**kwargs) -> None:
    dataset = get_resource(kwargs['test_info']['dataset'])
    run_test_dataset_search(dataset)


def test_build_subqueries(**kwargs) -> None:
    params = ASFSearchOptions(**get_resource(kwargs['test_info']['params']))
    expected = [
        ASFSearchOptions(**subquery) for subquery in get_resource(kwargs['test_info']['expected'])
    ]
    run_test_build_subqueries(params, expected)


def test_serialization(**args) -> None:
    test_info = args['test_info']
    product = get_resource(test_info.get('product'))
    results = get_resource(test_info.get('results'))
    search_opts = get_resource(test_info.get('searchOpts'))
    options = ASFSearchOptions(**search_opts if search_opts else {})

    run_test_serialization(product, results, options)


def test_notebook_examples(**args) -> None:
    test_info = args['test_info']
    notebook_file = test_info['notebook']
    path = os.path.join('examples', notebook_file)

    with open(path) as f:
        notebook = nbformat.read(f, as_version=4)
        ep = ExecutePreprocessor(timeout=600)
        try:
            assert ep.preprocess(notebook) is not None, f'Got empty notebook for {notebook_file}'
        except Exception as e:
            assert False, f'Failed executing {notebook_file}: {e}'


# Testing resource loading utilities


def safe_load_tuple(param):
    """
    loads a tuple from a list if a param is an object with key 'tuple'
    (Arbritrary constructor initialization is not supported by yaml.safe_load
    as a security measure)

    """
    if isinstance(param, Dict):
        if 'tuple' in param.keys():
            param = tuple(param['tuple'])

    return param


# def test_output_format(**args) -> None:
#     test_info = args['test_info']

#     products = get_resource(test_info['results'])
#     if not isinstance(products, List):
#         products = [products]
#     results = ASFSearchResults([as_ASFProduct({'meta': product['meta'], 'umm': product['umm']}, ASFSession()) for product in products])

#     run_test_output_format(results)


def test_keyword_aliasing_results(**args) -> None:
    test_info = args['test_info']

    opts = ASFSearchOptions(**test_info['params'])
    opts.maxResults = 250

    run_test_keyword_aliasing_results(opts)

def test_collection_attributes(**args) -> None:
    params = args['test_info']['params']
    expected_attributes = args['test_info']['expected_attributes']
    expect_failure = args['test_info'].get('expect_failure', False)
    session = args["config"].getoption("authenticated_session")
    run_test_collection_attributes(params, expected_attributes, session, expect_failure)

# Finds and loads file from yml_tests/Resouces/ if loaded field ends with .yml/yaml extension
def get_resource(yml_file):
    if isinstance(yml_file, str):
        if yml_file.endswith(('.yml', '.yaml')):
            base_path = pathlib.Path(__file__).parent.resolve()
            with open(os.path.join(base_path, 'yml_tests', 'Resources', yml_file), 'r') as f:
                try:
                    return yaml.safe_load(f)
                except yaml.YAMLError as exc:
                    print(exc)
    elif isinstance(yml_file, List):  # check if it's a list of yml files
        if len(yml_file) > 0:
            if isinstance(yml_file[0], str):
                if yml_file[0].endswith(('.yml', '.yaml')):
                    return [get_resource(file) for file in yml_file]

    return yml_file



from asf_search import ASFProduct, ALOS2Product
from math import sqrt, cos, sin, radians
from typing import List

import numpy as np

try:
    from ciso8601 import parse_datetime
except ImportError:
    from dateutil.parser import parse as parse_datetime

# WGS84 constants
a = 6378137
f = pow((1.0 - 1 / 298.257224), 2)
# Technically f is normally considered to just be that 298... part but this is all we ever use, so
# pre-calc and cache and call it all f anyhow


def calculate_perpendicular_baselines(reference: str, stack: List[ASFProduct]):
    isAlos2 = isinstance(stack[0], ALOS2Product)

    if not isAlos2:
        for product in stack:
            baselineProperties = product.baseline
            positionProperties = baselineProperties["stateVectors"]["positions"]

            if len(positionProperties.keys()) == 0:
                baselineProperties["noStateVectors"] = True
                continue
            if None in [
                positionProperties["prePositionTime"],
                positionProperties["postPositionTime"],
                positionProperties["prePosition"],
                positionProperties["postPosition"],
            ]:
                baselineProperties["noStateVectors"] = True
                continue

            asc_node_time = parse_datetime(
                baselineProperties["ascendingNodeTime"]
            ).timestamp()

            start = parse_datetime(product.properties["startTime"]).timestamp()
            end = parse_datetime(product.properties["stopTime"]).timestamp()
            center = start + ((end - start) / 2)
            baselineProperties["relative_start_time"] = start - asc_node_time
            baselineProperties["relative_center_time"] = center - asc_node_time
            baselineProperties["relative_end_time"] = end - asc_node_time

            t_pre = parse_datetime(positionProperties["prePositionTime"]).timestamp()
            t_post = parse_datetime(positionProperties["postPositionTime"]).timestamp()
            product.baseline["relative_sv_pre_time"] = t_pre - asc_node_time
            product.baseline["relative_sv_post_time"] = t_post - asc_node_time

    for product in stack:
        if product.properties["sceneName"] == reference:
            reference = product
            reference.properties["perpendicularBaseline"] = 0
            # Cache these values
            reference.baseline["granulePosition"] = get_granule_position(
                reference.properties["centerLat"], reference.properties["centerLon"]
            )
            break

    for secondary in stack:
        if secondary.baseline.get("noStateVectors"):
            secondary.properties["perpendicularBaseline"] = None
            continue

        if isAlos2:
            reference_shared_pos = reference.baseline["granulePosition"] + reference.baseline["stateVectors"]['position']
            reference_shared_vel = reference.baseline["stateVectors"]['velocity']
            secondary_shared_pos = secondary.baseline["stateVectors"]['position'] + get_granule_position(secondary.properties['center_lat'], secondary.properties['center_lon'])
        else:
            shared_rel_time = get_shared_sv_time(reference, secondary)
            reference_shared_pos = get_pos_at_rel_time(reference, shared_rel_time)
            reference_shared_vel = get_vel_at_rel_time(reference, shared_rel_time)
            secondary_shared_pos = get_pos_at_rel_time(secondary, shared_rel_time)

        # need to get sat pos and sat vel at center time
        reference.baseline["alongBeamVector"] = get_along_beam_vector(
            reference_shared_pos, reference.baseline["granulePosition"]
        )
        reference.baseline["upBeamVector"] = get_up_beam_vector(
            reference_shared_vel, reference.baseline["alongBeamVector"]
        )

        perpendicular_baseline = get_paired_granule_baseline(
            reference.baseline["granulePosition"],
            reference.baseline["upBeamVector"],
            secondary_shared_pos,
        )
        if abs(perpendicular_baseline) > 100000:
            perpendicular_baseline = None
        secondary.properties["perpendicularBaseline"] = perpendicular_baseline

    return stack


# Convert granule center lat/lon to fixed earth coordinates in meters using WGS84 ellipsoid.
def get_granule_position(scene_center_lat, scene_center_lon):
    lat = radians(float(scene_center_lat))
    lon = radians(float(scene_center_lon))
    coslat = cos(lat)  # This value gets used a couple times, cache it
    sinlat = sin(lat)  # This value gets used a couple times, cache it
    C = 1.0 / (sqrt(pow(coslat, 2) + f * pow(sinlat, 2)))
    S = f * C
    aC = a * C
    granule_position = np.array(
        [aC * coslat * cos(lon), aC * coslat * sin(lon), a * S * sinlat]
    )
    return granule_position


# Calculate along beam vector from sat pos and granule pos
def get_along_beam_vector(satellite_position, granule_position):
    along_beam_vector = np.subtract(satellite_position, granule_position)
    along_beam_vector = np.divide(
        along_beam_vector, np.linalg.norm(along_beam_vector)
    )  # normalize
    return along_beam_vector


# Calculate up beam vector from sat velocity and along beam vector
def get_up_beam_vector(satellite_velocity, along_beam_vector):
    up_beam_vector = np.cross(satellite_velocity, along_beam_vector)
    up_beam_vector = np.divide(
        up_beam_vector, np.linalg.norm(up_beam_vector)
    )  # normalize
    return up_beam_vector


# Calculate baseline between reference and paired granule
def get_paired_granule_baseline(
    reference_granule_position, reference_up_beam_vector, paired_satellite_position
):
    posd = np.subtract(paired_satellite_position, reference_granule_position)
    baseline = np.dot(reference_up_beam_vector, posd)
    return int(round(baseline))


# Find a relative orbit time covered by both granules' SVs
def get_shared_sv_time(reference, secondary):
    start = max(
        reference.baseline["relative_sv_pre_time"],
        secondary.baseline["relative_sv_pre_time"],
    )
    end = max(
        reference.baseline["relative_sv_post_time"],
        secondary.baseline["relative_sv_post_time"],
    )

    # Favor the start/end SV time of the reference so
    # we can use that SV directly without interpolation
    if start == reference.baseline["relative_sv_pre_time"]:
        return start
    if end == reference.baseline["relative_sv_post_time"]:
        return end

    return start


# Interpolate a position SV based on relative time
def get_pos_at_rel_time(granule: ASFProduct, relative_time):
    if relative_time == granule.baseline["relative_sv_pre_time"]:
        return granule.baseline["stateVectors"]["positions"]["prePosition"]
    if relative_time == granule.baseline["relative_sv_post_time"]:
        return granule.baseline["stateVectors"]["positions"]["postPosition"]

    duration = (
        granule.baseline["relative_sv_post_time"] - granule.baseline["relative_sv_pre_time"]
    )
    factor = (relative_time - granule.baseline["relative_sv_pre_time"]) / duration

    vec_a = granule.baseline["stateVectors"]["positions"]["prePosition"]
    vec_b = granule.baseline["stateVectors"]["positions"]["postPosition"]

    v = [
        interpolate(vec_a[0], vec_b[0], factor),
        interpolate(vec_a[1], vec_b[1], factor),
        interpolate(vec_a[2], vec_b[2], factor),
    ]

    return radius_fix(granule, v, relative_time)


# Interpolate a velocity SV based on relative time
def get_vel_at_rel_time(granule: ASFProduct, relative_time):
    velocityProperties = granule.baseline["stateVectors"]["velocities"]
    if relative_time == granule.baseline["relative_sv_pre_time"]:
        return velocityProperties["preVelocity"]
    if relative_time == granule.baseline["relative_sv_post_time"]:
        return velocityProperties["postVelocity"]

    duration = (
        granule.baseline["relative_sv_post_time"] - granule.baseline["relative_sv_pre_time"]
    )
    factor = (relative_time - granule.baseline["relative_sv_pre_time"]) / duration

    vec_a = velocityProperties["preVelocity"]
    vec_b = velocityProperties["postVelocity"]

    v = [
        interpolate(vec_a[0], vec_b[0], factor),
        interpolate(vec_a[1], vec_b[1], factor),
        interpolate(vec_a[2], vec_b[2], factor),
    ]

    return v


# convenience 1d linear interp
def interpolate(p0, p1, x):
    return (p0 * (1.0 - x)) + (p1 * x)


# Bump the provided sat pos out to a radius interpolated between the start and end sat pos vectors
def radius_fix(granule: ASFProduct, sat_pos, relative_time):
    positionProperties = granule.baseline["stateVectors"]["positions"]
    pre_l = np.linalg.norm(positionProperties["prePosition"])
    post_l = np.linalg.norm(positionProperties["postPosition"])
    sat_pos_l = np.linalg.norm(sat_pos)
    dt = relative_time - granule.baseline["relative_sv_pre_time"]
    new_l = pre_l + (post_l - pre_l) * dt / (
        granule.baseline["relative_sv_post_time"] - granule.baseline["relative_sv_pre_time"]
    )
    sat_pos[0] = sat_pos[0] * new_l / sat_pos_l
    sat_pos[1] = sat_pos[1] * new_l / sat_pos_l
    sat_pos[2] = sat_pos[2] * new_l / sat_pos_l
    return sat_pos



"""Various constants to be used in search and related functions,
provided as a convenience to help ensure sensible values."""

from .BEAMMODE import *  # noqa: F403 F401
from .FLIGHT_DIRECTION import *  # noqa: F403 F401
from .INSTRUMENT import *  # noqa: F403 F401
from .PLATFORM import *  # noqa: F403 F401
from .POLARIZATION import *  # noqa: F403 F401
from .PRODUCT_TYPE import *  # noqa: F403 F401
from .INTERNAL import *  # noqa: F403 F401
from .DATASET import *  # noqa: F403 F401
from .RANGE_BANDWIDTH import *  # noqa: F403 F401
from .PRODUCTION_CONFIGURATION import * # noqa: F403 F401



from asf_search.ASFSession import ASFSession
from asf_search.search.collection_attributes import get_searchable_attributes
import pytest

def run_test_collection_attributes(params: dict, expected_attributes: list[str], session: ASFSession, expect_failure: bool) -> None:
    if expect_failure:
        with pytest.raises(ValueError):
            actual_attributes = get_searchable_attributes(**params, session=session)
    else:
        actual_attributes = get_searchable_attributes(**params, session=session)
        assert sorted(expected_attributes) == sorted(actual_attributes.keys())



from typing import Dict
from asf_search import ASFProduct, ASFSession


class SIRCProduct(ASFProduct):
    """
    Dataset Documentation Page: https://eospso.nasa.gov/missions/spaceborne-imaging-radar-c
    """

    _base_properties = {
        **ASFProduct._base_properties,
        'groupID': {'path': ['AdditionalAttributes', ('Name', 'GROUP_ID'), 'Values', 0]},
        'md5sum': {'path': ['AdditionalAttributes', ('Name', 'MD5SUM'), 'Values', 0]},
        'pgeVersion': {'path': ['PGEVersionClass', 'PGEVersion']},
        'beamModeType': {'path': ['AdditionalAttributes', ('Name', 'BEAM_MODE_TYPE'), 'Values', 0]},
    }

    def __init__(self, args: Dict = {}, session: ASFSession = ASFSession()):
        super().__init__(args, session)



import copy
from asf_search.ASFSearchOptions import validators, ASFSearchOptions
from asf_search.ASFSearchOptions.config import config
from asf_search.ASFSearchOptions.validator_map import validate, validator_map
from pytest import raises


def run_test_validator_map_validate(key, value, output):
    if key not in list(validator_map.keys()):
        with raises(KeyError) as keyerror:
            validate(key, value)

        if key in [
            validator_key.lower()
            for validator_key in list(validator_map.keys())
            if key not in config.keys()
        ]:
            assert 'Did you mean' in str(keyerror.value)

        return

    assert validate(key, value) == output


def run_test_ASFSearchOptions_validator(validator_name, param, output, error):
    validator = getattr(validators, validator_name)

    if error is None:
        assert output == validator(param)
    else:
        with raises(ValueError) as e:
            validator(param)
        assert error in str(e.value)


def run_test_ASFSearchOptions(**kwargs):
    test_info = copy.copy(kwargs['test_info'])
    exception = test_info['exception']  # Can be "None" for don't.
    if 'expect_output' in test_info:
        expect_output = test_info.pop('expect_output')
    else:
        expect_output = {}

    # Take out anything that isn't supposed to reach the options object:
    del test_info['title']
    del test_info['exception']

    try:
        options_obj = ASFSearchOptions(**test_info)
    except (KeyError, ValueError) as e:
        assert (
            type(e).__name__ == exception
        ), f"ERROR: Didn't expect exception {type(e).__name__} to occur."
        return
    else:
        assert (
            exception is None
        ), f'ERROR: Expected exception {exception}, but SearchOptions never threw.'

    for key, val in expect_output.items():
        assert (
            getattr(options_obj, key) == val
        ), f"ERROR: options object param '{key}' should have value '{val}'. Got '{getattr(options_obj, key)}'."

    # test ASFSearchOptions.reset_search()
    options_obj.reset_search()

    assert (
        len([val for key, val in dict(options_obj).items() if key not in config.keys()]) == 0
    ), 'ERROR: ASFSearchOptions.reset() did not clear all non-default searchable params'

    for key, value in config.items():
        if test_info.get(key) is not None:
            assert (
                getattr(options_obj, key) == test_info[key]
            ), f"ERROR: User defined value '{test_info[key]}' for default param '{key}', but value was lost after ASFSearchOptions.reset()"
        else:
            assert (
                getattr(options_obj, key) == value
            ), f"ERROR: default param '{key}' left default by user changed, should have value '{val}'. Got '{getattr(options_obj, key)}'."



from .export_translators import ASFSearchResults_to_properties_list  # noqa: F401
from .csv import results_to_csv  # noqa: F401
from .metalink import results_to_metalink  # noqa: F401
from .kml import results_to_kml  # noqa: F401
from .jsonlite import results_to_jsonlite  # noqa: F401
from .jsonlite2 import results_to_jsonlite2  # noqa: F401
from .geojson import results_to_geojson  # noqa: F401
from .json import results_to_json   # noqa: F401



from copy import copy
from typing import Dict, Union

from asf_search import ASFSearchOptions, ASFSession, ASFStackableProduct
from asf_search.CMR.translate import try_parse_float, try_parse_int, try_round_float
from asf_search.constants import PRODUCT_TYPE

class ALOS2Product(ASFStackableProduct):
    """
    Used for ALOS Palsar and Avnir dataset products

    ASF Dataset Documentation Page: https://asf.alaska.edu/datasets/daac/alos-palsar/
    """

    _base_properties = {
        **ASFStackableProduct._base_properties,
        'frameNumber': {
            'path': ['AdditionalAttributes', ('Name', 'FRAME_NUMBER'), 'Values', 0],
            'cast': try_parse_int,
        },  # Sentinel and ALOS product alt for frameNumber (ESA_FRAME)
        'center_lat': {
            'path': ['AdditionalAttributes', ('Name', 'CENTER_LAT'), 'Values', 0],
            'cast': try_parse_float,
        },
        'center_lon': {
            'path': ['AdditionalAttributes', ('Name', 'CENTER_LON'), 'Values', 0],
            'cast': try_parse_float,
        },'faradayRotation': {
            'path': ['AdditionalAttributes', ('Name', 'FARADAY_ROTATION'), 'Values', 0],
            'cast': try_parse_float,
        },
        'offNadirAngle': {
            'path': ['AdditionalAttributes', ('Name', 'OFF_NADIR_ANGLE'), 'Values', 0],
            'cast': try_parse_float,
        },
        'bytes': {
            'path': ['DataGranule', 'ArchiveAndDistributionInformation', 0, 'SizeInBytes'],
            'cast': try_round_float,
        },
        'beamModeType': {'path': ['AdditionalAttributes', ('Name', 'BEAM_MODE_TYPE'), 'Values', 0]},
        'polarization': {
            'path': ['AdditionalAttributes', ('Name', 'POLARIZATION'), 'Values']
        },  # dual polarization is in list rather than a 'VV+VH' style format
    }

    baseline_type = ASFStackableProduct.BaselineCalcType.CALCULATED


    def __init__(self, args: Dict = {}, session: ASFSession = ASFSession()):
        super().__init__(args, session)
        self.properties.pop('md5sum')
        self.properties.pop('granuleType')
        self.properties.pop('processingLevel')
        
        self.baseline = self.get_baseline_calc_properties()

    def get_baseline_calc_properties(self) -> Dict:
        """
        :returns properties required for SLC baseline stack calculations
        """
        return {'stateVectors': self.get_state_vectors()}

    def get_stack_opts(self, opts: ASFSearchOptions = None) -> ASFSearchOptions:
        """
        Returns the search options asf-search will use internally
        to build an SLC baseline stack from

        :param opts: additional criteria for limiting
        :returns ASFSearchOptions used for build Sentinel-1 SLC Stack
        """
        stack_opts = ASFSearchOptions() if opts is None else copy(opts)

        stack_opts.beamMode = [self.properties['beamModeType']]
        stack_opts.flightDirection = self.properties['flightDirection']
        stack_opts.relativeOrbit = [int(self.properties['pathNumber'])]  # path
        stack_opts.dataset = 'ALOS-2'
        
        if any(e in ['HH', 'HH+HV'] for e in self.properties['polarization']):
            stack_opts.polarization = ['HH', 'HH+HV']
        else:
            stack_opts.polarization = ['VV', 'VV+VH']

        stack_opts.intersectsWith = self.centroid().wkt

        return stack_opts

    def get_state_vectors(self) -> Dict:
        """
        Used in spatio-temporal perpendicular baseline calculations for non-pre-calculated stacks

        :returns dictionary of pre/post positions, velocities, and times"""

        position = [
            float(val)
            for val in self.umm_get(
                self.umm, 'AdditionalAttributes', ('Name', 'SV_POSITION'), 'Values'
            )
        ]
        velocity = [
            float(val)
            for val in self.umm_get(
                self.umm, 'AdditionalAttributes', ('Name', 'SV_VELOCITY'), 'Values'
            )
        ]

        return dict(position=position, velocity=velocity)

    def is_valid_reference(self):
        return self.has_baseline()



from dataclasses import dataclass
from typing import Optional
from asf_search.CMR.datasets import collections_by_processing_level
from asf_search.ASFSession import ASFSession

from asf_search.exceptions import ASFSearchError, CMRError


@dataclass(frozen=True)
class AdditionalAttribute:
    """Wrapper dataclass around CMR Additional Attributes"""

    name: str
    """The `Name` of the additional attribute in CMR"""
    data_type: str
    """The `DataType` of the additional attribute in CMR"""
    description: str
    """The `Description` of the additional attribute in CMR"""


def get_searchable_attributes(
    shortName: Optional[str] = None,
    conceptID: Optional[str] = None,
    processingLevel: Optional[str] = None,
    session: ASFSession = ASFSession(),
) -> dict[str, AdditionalAttribute]:
    """Using a provided processingLevel, collection shortName, or conceptID query CMR's `/collections` endpoint and
    return a dictionary of additional attributes mapping the attribute's name to the additional attribute entry in CMR

    ``` python
    from pprint import pp
    SLCRcord = asf.get_searchable_attributes(processingLevel='SLC')
    pp(SLCRcord.additionalAttributes)
    ```
    """
    query_data = None
    method = None

    if shortName is not None:
        method = {'type': 'shortName', 'value': shortName}
        query_data = [('shortName', shortName)]
    elif conceptID is not None:
        query_data = [('concept-id', conceptID)]
        method = {'type': 'conceptID', 'value': conceptID}
    elif processingLevel is not None:
        method = {'type': 'processingLevel', 'value': processingLevel}
        query_data = _get_concept_ids_for_processing_level(processingLevel)
    else:
        raise ValueError(
            'Error: `get_collection_searchable_attributes()` expects `shortName`, `conceptID`, or `processingLevel`'
        )

    cmr_response = _query_cmr(session=session, query_data=query_data, method=method)

    if 'errors' in cmr_response:
        raise ValueError(f"CMR responded with an error. Original error(s): {' '.join(cmr_response['errors'])}")
    if len(cmr_response['items']) == 0:
        raise ValueError(
            f'Error: no collections found in CMR for given parameter `{method["type"]}`: "{method["value"]}" '
        )

    additionalAttributes = {}

    for entry in cmr_response['items']:
        umm = entry['umm']
        attributes = umm.get('AdditionalAttributes')
        if attributes is not None:
            for attribute in attributes:
                additionalAttributes[attribute.get('Name')] = AdditionalAttribute(
                    name=attribute.get('Name'),
                    description=attribute.get('Description'),
                    data_type=attribute.get('DataType'),
                )

    return additionalAttributes


def _get_concept_ids_for_processing_level(processing_level: str):
    collections = collections_by_processing_level.get(processing_level)
    if collections is None:
        raise ValueError(f'asf-search is missing concept-id aliases for processing level "{processing_level}". Please use `shortName` or `conceptID')
    return [('concept-id[]', collection) for collection in collections]


def _query_cmr(session: ASFSession, query_data: list[tuple[str, str]], method: dict) -> dict:
    url = 'https://cmr.earthdata.nasa.gov/search/collections.umm_json'

    response = session.post(url=url, data=query_data)

    try:
        return response.json()
    except Exception as exc:
        raise ASFSearchError(
            f'Failed to find collection attributes for {method["type"]} "{method["value"]}". original exception: {str(exc)}'
        )



from importlib.metadata import PackageNotFoundError, version

## Setup logging now, so it's available if __version__ fails:
import logging
ASF_LOGGER = logging.getLogger(__name__)
# Add null handle so we do nothing by default. It's up to whatever
# imports us, if they want logging.
ASF_LOGGER.addHandler(logging.NullHandler())

try:
    __version__ = version(__name__)
except PackageNotFoundError as e:
    msg = str(
        "package is not installed!\n"
        "Install in editable/develop mode via (from the top of this repo):\n"
        "   python3 -m pip install -e .\n"
        "Or, to just get the version number use:\n"
        "   python setup.py --version"
    )
    print(msg)
    ASF_LOGGER.exception(msg)  # type: ignore # noqa: F821
    raise PackageNotFoundError(
        "Install with 'python3 -m pip install -e .' to use"
    ) from e

ASF_LOGGER = logging.getLogger(__name__)
# Add null handle so we do nothing by default. It's up to whatever
# imports us, if they want logging.
ASF_LOGGER.addHandler(logging.NullHandler())

from .ASFSession import ASFSession  # noqa: F401, E402
from .ASFProduct import ASFProduct  # noqa: F401 E402
from .ASFStackableProduct import ASFStackableProduct  # noqa: F401 E402
from .ASFSearchResults import ASFSearchResults  # noqa: F401 E402
from .ASFSearchOptions import ASFSearchOptions, validators  # noqa: F401 E402
from .Products import *  # noqa: F403 F401 E402
from .exceptions import *  # noqa: F403 F401 E402
from .constants import (  # noqa: F401 E402
    BEAMMODE,  # noqa: F401 E402
    FLIGHT_DIRECTION,  # noqa: F401 E402
    INSTRUMENT,  # noqa: F401 E402
    PLATFORM,  # noqa: F401 E402
    POLARIZATION,  # noqa: F401 E402
    PRODUCT_TYPE,  # noqa: F401 E402
    INTERNAL,  # noqa: F401 E402
    DATASET,  # noqa: F401 E402
    RANGE_BANDWIDTH,  # noqa: F401 E402,
    PRODUCTION_CONFIGURATION,  # noqa: F401 E402
)
from .health import *  # noqa: F403 F401 E402
from .search import *  # noqa: F403 F401 E402
from .download import *  # noqa: F403 F401 E402
from .CMR import *  # noqa: F403 F401 E402
from .baseline import *  # noqa: F403 F401 E402
from .WKT import validate_wkt  # noqa: F401 E402
from .export import *  # noqa: F403 F401 E402
from .Pair import Pair  # noqa:  F401, E402

REPORT_ERRORS = True
"""Enables automatic search error reporting to ASF, send any questions to uso@asf.alaska.edu"""



SENTINEL1 = 'SENTINEL-1'
OPERA_S1 = 'OPERA-S1'
OPERA_S1_CALVAL = 'OPERA-S1-CALVAL'
SLC_BURST = 'SLC-BURST'
ALOS_PALSAR = 'ALOS PALSAR'
ALOS_AVNIR_2 = 'ALOS AVNIR-2'
ALOS_2 = 'ALOS-2'
SIRC = 'SIR-C'
ARIA_S1_GUNW = 'ARIA S1 GUNW'
SMAP = 'SMAP'
UAVSAR = 'UAVSAR'
RADARSAT_1 = 'RADARSAT-1'
ERS = 'ERS'
JERS_1 = 'JERS-1'
AIRSAR = 'AIRSAR'
SEASAT = 'SEASAT'
NISAR = 'NISAR'
"""NISAR provides L and S-band SAR data to measure Earth's changing ecosystems,
 dynamic surfaces, and ice masses with 12-day regularity
 on ascending and descending passes."""



from numbers import Number
from asf_search import ASFProduct, ASFSearchOptions
from asf_search import ASFSession
# from asf_search.CMR.translate import get

from tenacity import retry, retry_if_exception_type, stop_after_attempt
from asf_search import ASF_LOGGER
from asf_search.CMR.subquery import build_subqueries
from asf_search.CMR.translate import try_parse_date
from asf_search.constants import INTERNAL
from asf_search.exceptions import ASFSearchError
from asf_search.search import search
from asf_search.ASFSearchResults import ASFSearchResults
from asf_search.CMR import dataset_collections
from pytest import raises
from typing import List
import requests
import requests_mock

from asf_search.search.search_generator import as_ASFProduct, preprocess_opts

SEARCHAPI_URL = 'https://api.daac.asf.alaska.edu'
SEARCHAPI_ENDPOINT = '/services/search/param?'


def run_test_ASFSearchResults(search_resp):
    search_results = ASFSearchResults(
        [as_ASFProduct(product, ASFSession()) for product in search_resp]
    )

    assert len(search_results) == len(search_resp)
    assert search_results.geojson()['type'] == 'FeatureCollection'

    for idx, feature in enumerate(search_results):
        # temporal and perpendicular baseline values are calculated post-search,
        # so there's no instance where they'll be returned in a CMR search
        search_resp[idx]['properties'].pop('temporalBaseline', None)
        search_resp[idx]['properties'].pop('perpendicularBaseline', None)

        assert feature.geojson()['geometry'] == search_resp[idx]['geometry']
        for key, item in feature.geojson()['properties'].items():
            if key == 'esaFrame':
                assert search_resp[idx]['properties']['frameNumber'] == item
            elif 'esaFrame' in feature.geojson()['properties'].keys() and key == 'frameNumber':
                continue
            elif key in ['stopTime', 'startTime', 'processingDate']:
                assert try_parse_date(item) == try_parse_date(search_resp[idx]['properties'][key])
            elif search_resp[idx]['properties'].get(key) is not None and item is not None:
                assert item == search_resp[idx]['properties'][key]


def run_test_search(search_parameters, answer):
    with requests_mock.Mocker() as m:
        m.post(
            f'https://{INTERNAL.CMR_HOST}{INTERNAL.CMR_GRANULE_PATH}',
            json={'items': answer, 'hits': len(answer)},
        )
        response = search(**search_parameters)

        if search_parameters.get('maxResults', False):
            assert len(response) == search_parameters['maxResults']

        assert len(response) == len(answer)
        # assert(response.geojson()["features"] == answer)


def run_test_search_http_error(search_parameters, status_code: Number, report: str):
    if not len(search_parameters.keys()):
        with requests_mock.Mocker() as m:
            m.register_uri(
                'POST',
                f'https://{INTERNAL.CMR_HOST}{INTERNAL.CMR_GRANULE_PATH}',
                status_code=status_code,
                json={'errors': {'report': report}},
            )
            m.register_uri('POST', 'https://search-error-report.asf.alaska.edu/', real_http=True)
            searchOptions = ASFSearchOptions(**search_parameters)
            with raises(ASFSearchError):
                search(opts=searchOptions)
            return

    # If we're not doing an empty search we want to fire off one real query to CMR, then interrupt it with an error
    # We can tell a search isn't the first one by checking if 'CMR-Search-After' has been set
    def custom_matcher(request: requests.Request):
        if 'CMR-Search-After' in request.headers.keys():
            resp = requests.Response()
            resp.status_code = 200
            return resp
        return None

    with requests_mock.Mocker() as m:
        m.register_uri(
            'POST',
            f'https://{INTERNAL.CMR_HOST}{INTERNAL.CMR_GRANULE_PATH}',
            real_http=True,
        )
        m.register_uri(
            'POST',
            f'https://{INTERNAL.CMR_HOST}{INTERNAL.CMR_GRANULE_PATH}',
            additional_matcher=custom_matcher,
            status_code=status_code,
            json={'errors': {'report': report}},
        )
        m.register_uri('POST', 'https://search-error-report.asf.alaska.edu/', real_http=True)

        search_parameters['maxResults'] = INTERNAL.CMR_PAGE_SIZE + 1
        searchOptions = ASFSearchOptions(**search_parameters)

        with raises(ASFSearchError):
            search(opts=searchOptions)


def run_test_dataset_search(datasets: List):
    if any(dataset for dataset in datasets if dataset_collections.get(dataset) is None):
        with raises(ValueError):
            search(dataset=datasets, maxResults=1)
    else:
        for dataset in datasets:
            valid_shortnames = list(dataset_collections.get(dataset))

            response = search(dataset=dataset, maxResults=250)

            # Get collection shortName of all granules
            shortNames = list(
                set(
                    [
                        shortName
                        for product in response
                        if (
                            shortName := ASFProduct.umm_get(
                                product.umm, 'CollectionReference', 'ShortName'
                            )
                        )
                        is not None
                    ]
                )
            )

            # and check that results are limited to the expected datasets by their shortname
            for shortName in shortNames:
                assert shortName in valid_shortnames


def run_test_build_subqueries(params: ASFSearchOptions, expected: List):
    # mainly for getting platform aliases
    preprocess_opts(params)
    actual = build_subqueries(params)
    for a, b in zip(actual, expected):
        for key, actual_val in a:
            expected_val = getattr(b, key)
            if isinstance(actual_val, list):
                if key == 'cmr_keywords':
                    for idx, key_value_pair in enumerate(actual_val):
                        assert key_value_pair == expected_val[idx]
                else:
                    if len(actual_val) > 0:  # ASFSearchOptions leaves empty lists as None
                        expected_set = set(expected_val)
                        actual_set = set(actual_val)

                        difference = expected_set.symmetric_difference(actual_set)
                        assert (
                            len(difference) == 0
                        ), f'Found {len(difference)} missing entries for subquery generated keyword: "{key}"\n{list(difference)}'
            else:
                assert actual_val == expected_val


def run_test_keyword_aliasing_results(params: ASFSearchOptions):
    module_response = search(opts=params)

    try:
        api_response = query_endpoint(dict(params))
    except requests.ReadTimeout:
        ASF_LOGGER.warn(f'SearchAPI timed out, skipping test for params {str(params)}')
        return

    api_results = api_response['results']

    api_dict = {product['granuleName']: True for product in api_results}

    for product in module_response:
        sceneName = product.properties['sceneName']
        assert api_dict.get(
            sceneName, False
        ), f'Found unexpected scene in asf-search module results, {sceneName}\{dict(params)}'


@retry(
    stop=stop_after_attempt(3),
    retry=retry_if_exception_type(requests.HTTPError),
    reraise=True,
)
def query_endpoint(params):
    response = requests.post(
        url=SEARCHAPI_URL + SEARCHAPI_ENDPOINT, data={**params, 'output': 'jsonlite'}
    )
    response.raise_for_status()

    return response.json()



from typing import Optional, Type
from asf_search.baseline.stack import get_baseline_from_stack
from asf_search import ASF_LOGGER
from copy import copy

from asf_search.search import search, product_search
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.ASFSearchResults import ASFSearchResults
from asf_search import ASFProduct, ARIAS1GUNWProduct
from asf_search.constants import PLATFORM, DATASET
from asf_search.exceptions import ASFSearchError


precalc_platforms = [
    PLATFORM.ALOS,
    PLATFORM.RADARSAT,
    PLATFORM.ERS1,
    PLATFORM.ERS2,
    PLATFORM.JERS,
]


def stack_from_product(
    reference: ASFProduct,
    opts: ASFSearchOptions = None,
    ASFProductSubclass: Type[ASFProduct] = None,
) -> ASFSearchResults:
    """
    Finds a baseline stack from a reference ASFProduct

    Parameters
    ----------
    reference:
        Reference scene to base the stack on,
        and from which to calculate perpendicular/temporal baselines
    opts:
        An ASFSearchOptions object describing the search parameters to be used.
        Search parameters specified outside this object will override in event of a conflict.
    ASFProductSubclass:
        An ASFProduct subclass constructor.

    Returns
    -------
    `asf_search.ASFSearchResults`
        list of search results of subclass ASFProduct or of provided ASFProductSubclass
    """

    opts = ASFSearchOptions() if opts is None else copy(opts)

    opts.merge_args(**dict(reference.get_stack_opts()))

    stack = search(opts=opts)

    is_complete = stack.searchComplete

    if ASFProductSubclass is not None:
        _cast_results_to_subclass(stack, ASFProductSubclass)

    stack, warnings = get_baseline_from_stack(reference=reference, stack=stack)

    _post_process_stack(stack, warnings, is_complete)


    return stack


def stack_from_id(
    reference_id: str,
    opts: Optional[ASFSearchOptions] = None,
    useSubclass: Optional[Type[ASFProduct]] = None,
) -> ASFSearchResults:
    """
    Finds a baseline stack from a reference product ID

    Parameters
    ----------
    reference_id:
        Reference product to base the stack from,
        and from which to calculate perpendicular/temporal baselines
    opts:
        An ASFSearchOptions object describing the search parameters to be used.
        Search parameters specified outside this object will override in event of a conflict.
    ASFProductSubclass:
        An ASFProduct subclass constructor.

    Returns
    -------
    `asf_search.ASFSearchResults`
        list of search results of subclass ASFProduct or of provided ASFProductSubclass
    """

    opts = ASFSearchOptions() if opts is None else copy(opts)

    if opts.dataset is not None and DATASET.ARIA_S1_GUNW in opts.dataset:
        reference_results = ARIAS1GUNWProduct.get_aria_groups_for_frame(reference_id)

        if len(reference_results) == 0:
            reference = None
        else:
            reference = reference_results[0]
        
        stack, warnings = get_baseline_from_stack(reference=reference, stack=reference_results)
        _post_process_stack(stack, warnings, reference_results.searchComplete)

        return stack
    else:
        reference_results = product_search(product_list=reference_id, opts=opts)

    if len(reference_results) <= 0:
        raise ASFSearchError(f'Reference product not found: {reference_id}')
    reference = reference_results[0]

    if useSubclass is not None:
        reference = _cast_to_subclass(reference, useSubclass)

    return reference.stack(opts=opts, useSubclass=useSubclass)


def _cast_results_to_subclass(stack: ASFSearchResults, ASFProductSubclass: Type[ASFProduct]):
    """
    Converts results from default ASFProduct subclasses to custom ones
    """
    for idx, product in enumerate(stack):
        stack[idx] = _cast_to_subclass(product, ASFProductSubclass)


def _cast_to_subclass(product: ASFProduct, subclass: Type[ASFProduct]) -> ASFProduct:
    """
    Casts this ASFProduct object as a new object of return type subclass.

    example:
    ```
    class MyCustomClass(ASFProduct):
        _base_properties = {
        **ASFProduct._base_properties,
        'some_unique_property': {'path': ['AdditionalAttributes', 'UNIQUE_PROPERTY', ...]}
        }

    # subclass as constructor
    customReference = reference.cast_to_subclass(MyCustomClass)
    print(customReference.properties['some_unique_property'])
    ```

    :param subclass: The ASFProduct subclass constructor to call on the product
    :returns return product as `ASFProduct` subclass
    """

    try:
        if isinstance(subclass, type(ASFProduct)):
            return subclass(
                args={'umm': product.umm, 'meta': product.meta}, session=product.session
            )
    except Exception as e:
        raise ValueError(f'Unable to use provided subclass {type(subclass)}, \nError Message: {e}')

    raise ValueError(f'Expected ASFProduct subclass constructor, got {type(subclass)}')

def _post_process_stack(stack: ASFSearchResults, warnings: list, is_complete: bool):
    """Marks whether the search completed gathering results, logs any warnings, and sorts stack by temporal baseline"""
    stack.searchComplete = is_complete  # preserve final outcome of earlier search()
    for warning in warnings:
        ASF_LOGGER.warning(f'{warning}')
    stack.sort(key=lambda product: product.properties['temporalBaseline'])



import time
from typing import Literal, Union, Sequence, Tuple
from copy import copy
import datetime

from asf_search import ASF_LOGGER, ASFSearchResults
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.exceptions import ASFSearchError
from asf_search.search.search_generator import search_generator


def search(
    absoluteOrbit: Union[
        int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]
    ] = None,
    asfFrame: Union[
        int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]
    ] = None,
    beamMode: Union[str, Sequence[str]] = None,
    beamSwath: Union[str, Sequence[str]] = None,
    campaign: Union[str, Sequence[str]] = None,
    maxDoppler: float = None,
    minDoppler: float = None,
    end: Union[datetime.datetime, str] = None,
    maxFaradayRotation: float = None,
    minFaradayRotation: float = None,
    flightDirection: str = None,
    flightLine: str = None,
    frame: Union[int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]] = None,
    granule_list: Union[str, Sequence[str]] = None,
    groupID: Union[str, Sequence[str]] = None,
    insarStackId: str = None,
    instrument: Union[str, Sequence[str]] = None,
    intersectsWith: str = None,
    lookDirection: Union[str, Sequence[str]] = None,
    offNadirAngle: Union[
        float, Tuple[float, float], Sequence[Union[float, Tuple[float, float]]]
    ] = None,
    platform: Union[str, Sequence[str]] = None,
    polarization: Union[str, Sequence[str]] = None,
    processingDate: Union[datetime.datetime, str] = None,
    processingLevel: Union[str, Sequence[str]] = None,
    product_list: Union[str, Sequence[str]] = None,
    relativeOrbit: Union[
        int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]
    ] = None,
    season: Tuple[int, int] = None,
    start: Union[datetime.datetime, str] = None,
    absoluteBurstID: Union[int, Sequence[int]] = None,
    relativeBurstID: Union[int, Sequence[int]] = None,
    fullBurstID: Union[str, Sequence[str]] = None,
    temporalBaselineDays: Union[str, Sequence[str]] = None,
    operaBurstID: Union[str, Sequence[str]] = None,
    frameCoverage: Literal['FULL', 'PARTIAL'] = None,
    mainBandPolarization: Union[str, Sequence[str]] = None,
    sideBandPolarization: Union[str, Sequence[str]] = None,
    rangeBandwidth: Union[str, Sequence[str]] = None,
    jointObservation: bool = None,
    productionConfiguration: Union[Literal["PR", "UR"], Sequence[Literal["PR", "UR"]]] = None,
    dataset: Union[str, Sequence[str]] = None,
    collections: Union[str, Sequence[str]] = None,
    shortName: Union[str, Sequence[str]] = None,
    cmr_keywords: Union[Tuple[str, str], Sequence[Tuple[str, str]]] = None,
    maxResults: int = None,
    opts: ASFSearchOptions = None,
) -> ASFSearchResults:
    """
    Performs a generic search against the Central Metadata Repository (CMR),
    returning all results in a single list.
    (For accessing results page by page see `asf_search.search_generator()`)

    Accepts a number of search parameters, and/or an ASFSearchOptions object.
    If an ASFSearchOptions object is provided as well as other specific parameters,
    the two sets of options will be merged, preferring the specific keyword arguments.

    Parameters
    ----------
    absoluteOrbit:
        For ALOS, ERS-1, ERS-2, JERS-1, and RADARSAT-1, Sentinel-1A, Sentinel-1B
        this value corresponds to the orbit count within the orbit cycle.
        For UAVSAR it is the Flight ID.
    asfFrame:
        This is primarily an ASF / JAXA frame reference. However,
        some platforms use other conventions. See ‘frame’ for ESA-centric frame searches.
    beamMode:
        The beam mode used to acquire the data.
    beamSwath:
        Encompasses a look angle and beam mode.
    campaign:
        For UAVSAR and AIRSAR data collections only. Search by general location,
        site description, or data grouping as supplied by flight agency or project.
    maxDoppler:
        Doppler provides an indication of how much the look direction deviates
        from the ideal perpendicular flight direction acquisition.
    minDoppler:
        Doppler provides an indication of how much the look direction deviates
        from the ideal perpendicular flight direction acquisition.
    end:
        End date of data acquisition. Supports timestamps
        as well as natural language such as "3 weeks ago"
    maxFaradayRotation:
        Rotation of the polarization plane of
        the radar signal impacts imagery, as HH and HV signals become mixed.
    minFaradayRotation:
        Rotation of the polarization plane of
        the radar signal impacts imagery, as HH and HV signals become mixed.
    flightDirection:
        Satellite orbit direction during data acquisition
    flightLine:
        Specify a flightline for UAVSAR or AIRSAR.
    frame:
        ESA-referenced frames are offered to give users a universal framing convention.
        Each ESA frame has a corresponding ASF frame assigned. See also: asfframe
    granule_list:
        List of specific granules.
        Search results may include several products per granule name.
    groupID:
        Identifier used to find products considered to
        be of the same scene but having different granule names
    insarStackId:
        Identifier used to find products of the same InSAR stack
    instrument:
        The instrument used to acquire the data. See also: platform
    intersectsWith:
        Search by polygon, linestring,
        or point defined in 2D Well-Known Text (WKT)
    lookDirection:
        Left or right look direction during data acquisition
    offNadirAngle:
        Off-nadir angles for ALOS PALSAR
    platform:
        Remote sensing platform that acquired the data.
        Platforms that work together, such as Sentinel-1A/1B and ERS-1/2
        have multi-platform aliases available. See also: instrument
    polarization:
        A property of SAR electromagnetic waves
        that can be used to extract meaningful information about surface properties of the earth.
    processingDate:
        Used to find data that has been processed at ASF since a given
        time and date. Supports timestamps as well as natural language such as "3 weeks ago"
    processingLevel:
        Level to which the data has been processed
    product_list:
        List of specific products.
        Guaranteed to be at most one product per product name.
    relativeOrbit:
        Path or track of satellite during data acquisition.
        For UAVSAR it is the Line ID.
    season:
        Start and end day of year for desired seasonal range.
        This option is used in conjunction with start/end to specify a seasonal range
        within an overall date range.
    start:
        Start date of data acquisition.
        Supports timestamps as well as natural language such as "3 weeks ago"
    collections:
        List of collections (concept-ids) to limit search to
    temporalBaselineDays:
        List of temporal baselines,
        used for Sentinel-1 Interferogram (BETA)
    maxResults:
        The maximum number of results to be returned by the search
    opts:
        An ASFSearchOptions object describing the search parameters to be used.
        Search parameters specified outside this object will override in event of a conflict.

    Returns
    -------
    `asf_search.ASFSearchResults` (list of search results of subclass ASFProduct)
    """
    kwargs = locals()
    data = dict((k, v) for k, v in kwargs.items() if k not in ['host', 'opts'] and v is not None)

    opts = ASFSearchOptions() if opts is None else copy(opts)
    opts.merge_args(**data)

    results = ASFSearchResults([])

    # The last page will be marked as complete if results sucessful
    perf = time.time()
    for page in search_generator(opts=opts):
        ASF_LOGGER.debug(f'Page Time Elapsed {time.time() - perf}')
        results.extend(page)
        results.searchComplete = page.searchComplete
        results.searchOptions = page.searchOptions
        perf = time.time()


    if not results.searchComplete:
        msg = (
            'Results may be incomplete due to a search error. '
            'See ASF_LOGGER logging for more details.'
        )

        ASF_LOGGER.error(msg)
        

    try:
        results.sort(key=lambda p: p.get_sort_keys(), reverse=True)
    except TypeError as exc:
        ASF_LOGGER.warning(f'Failed to sort final results, leaving results unsorted. Reason: {exc}')

    return results



import warnings
import json

from .validator_map import validator_map, validate
from .config import config
from asf_search import ASF_LOGGER


class ASFSearchOptions:
    def __init__(self, **kwargs):
        """
        Initialize the object, creating the list of attributes
        based on the contents of validator_map, and assign them based on kwargs

        :param kwargs: any search options to be set immediately
        """
        # init the built in attrs:
        for key in validator_map:
            self.__setattr__(key, None)

        # Apply any parameters passed in:
        for key, value in kwargs.items():
            self.__setattr__(key, value)

    def __setattr__(self, key, value):
        """
        Set a search option, restricting to the keys in validator_map only,
        and applying validation to the value before setting

        :param key: the name of the option to be set
        :param value: the value to which to set the named option
        """
        # self.* calls custom __setattr__ method, creating inf loop. Use super().*
        # Let values always be None, even if their validator doesn't agree. Used to delete them too:
        if key in validator_map:
            if value is None:  # always maintain config on required fields
                if key in config:
                    super().__setattr__(key, config[key])
                else:
                    super().__setattr__(key, None)
            else:
                super().__setattr__(key, validate(key, value))
        else:
            msg = f"key '{key}' is not a valid search option (setattr)"
            ASF_LOGGER.error(msg)
            raise KeyError(msg)

    def __delattr__(self, item):
        """
        Clear a search option by setting its value to None

        :param item: the name of the option to clear
        """
        if item in validator_map:
            self.__setattr__(item, None)
        else:
            msg = f"key '{item}' is not a valid search option (delattr)"
            ASF_LOGGER.error(msg)
            raise KeyError(msg)

    def __iter__(self):
        """
        Filters search parameters, only returning populated fields. Used when casting to a dict.
        """

        for key in validator_map:
            if not self._is_val_default(key):
                value = self.__getattribute__(key)
                yield key, value

    def __str__(self):
        """
        What to display if `print(opts)` is called.
        """
        return json.dumps(dict(self), indent=4, default=str)

    # Default is set to '...', since 'None' is a very valid value here
    def pop(self, key, default=...):
        """
        Removes 'key' from self and returns it's value. Throws KeyError if doesn't exist

        :param key: name of key to return value of, and delete
        """
        if key not in validator_map:
            msg = f"key '{key}' is not a valid key for ASFSearchOptions. (pop)"
            ASF_LOGGER.error(msg)
            raise KeyError(msg)

        if self._is_val_default(key):
            if default != ...:
                return default
            msg = f"key '{key}' is set to empty/None. (pop)"
            ASF_LOGGER.error(msg)
            raise KeyError(msg)

        # Success, delete and return it:
        val = getattr(self, key)
        self.__delattr__(key)
        return val

    def reset_search(self):
        """
        Resets all populated search options, excluding config options (host, session, etc)
        """
        for key, _ in self:
            if key not in config:
                super().__setattr__(key, None)

    def merge_args(self, **kwargs) -> None:
        """
        Merges all keyword args into this ASFSearchOptions object.
        Emits a warning for any options that are over-written by the operation.

        :param kwargs: The search options to merge into the object
        :return: None
        """
        for key in kwargs:
            # Spit out warning if the value is something other than the default:
            if not self._is_val_default(key):
                msg = (
                    'While merging search options, '
                    f'existing option {key}:{getattr(self, key, None)} '
                    f'overwritten by kwarg with value {kwargs[key]}'
                )
                ASF_LOGGER.warning(msg)
                warnings.warn(msg)
            self.__setattr__(key, kwargs[key])

    def _is_val_default(self, key) -> bool:
        """
        Returns bool on if the key's current value is the same as it's default value

        :param key: The key to check
        :return: bool
        """
        default_val = config[key] if key in config else None
        current_val = getattr(self, key, None)
        return current_val == default_val



from datetime import datetime, timedelta, timezone
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.search import product_search
from asf_search import Pair
import numpy as np


def test_make_s1_pairs():

    args = ASFSearchOptions(
        **{"start": '2022-02-10', "end": '2022-07-01'}
    )

    granule_product = product_search('S1A_IW_SLC__1SDV_20220215T225119_20220215T225146_041930_04FE2E_9252-SLC')[0]
    granule_stack = granule_product.stack(args)
    granule_pair = Pair(granule_product, granule_stack[1])
    assert granule_pair.ref.properties['sceneName'] == "S1A_IW_SLC__1SDV_20220215T225119_20220215T225146_041930_04FE2E_9252"
    assert granule_pair.sec.properties['sceneName'] == "S1A_IW_SLC__1SDV_20220227T225119_20220227T225146_042105_050431_987E"
    assert granule_pair.ref_time == datetime(2022, 2, 15, 22, 51, 19, tzinfo=timezone.utc)
    assert granule_pair.sec_time == datetime(2022, 2, 27, 22, 51, 19, tzinfo=timezone.utc)
    assert granule_pair.perpendicular_baseline == -15
    assert granule_pair.temporal_baseline == timedelta(days=12)
    assert np.floor(granule_pair.estimate_s1_mean_coherence()) == 18.0

    burst_product = product_search('S1_181296_IW1_20220219T125501_VV_10AF-BURST')[0]
    burst_stack = burst_product.stack(args)
    burst_pair = Pair(burst_product, burst_stack[1])
    assert burst_pair.ref.properties['sceneName'] == "S1_181296_IW1_20220219T125501_VV_10AF-BURST"
    assert burst_pair.sec.properties['sceneName'] == "S1_181296_IW1_20220303T125501_VV_F03A-BURST"
    assert burst_pair.ref_time == datetime(2022, 2, 19, 12, 55, 3, tzinfo=timezone.utc)
    assert burst_pair.sec_time == datetime(2022, 3, 3, 12, 55, 2, tzinfo=timezone.utc)
    assert burst_pair.perpendicular_baseline == -75
    assert burst_pair.temporal_baseline == timedelta(days=12)
    assert np.floor(burst_pair.estimate_s1_mean_coherence()) == 52.0



HH = 'HH'
VV = 'VV'
VV_VH = 'VV+VH'
HH_HV = 'HH+HV'
DUAL_HH = 'DUAL HH'
DUAL_VV = 'DUAL VV'
DUAL_HV = 'DUAL HV'
DUAL_VH = 'DUAL VH'
HH_3SCAN = 'HH 3SCAN'
HH_4SCAN = 'HH 4SCAN'
HH_5SCAN = 'HH 5SCAN'
QUAD = 'quadrature'
HH_VV = 'HH+VV'
HH_HV_VH_VV = 'HH+HV+VH+VV'
FULL = 'full'
UNKNOWN = 'UNKNOWN'
# NISAR
LH_LV="LH+LV"
RH_RV="RH+RV"



import argparse
import pytest
import os
from asf_search.ASFSession import ASFSession
from getpass import getpass

def string_to_session(user_input: str) -> ASFSession:
    session = ASFSession()

    if user_input is not None and len(user_input):
        session.auth_with_token(user_input)

    return session

def set_should_auth_session(user_input: str) -> ASFSession:
    should_auth = string_to_bool(user_input)
    session = ASFSession()
    if should_auth:
        if (token:=os.environ.get('EDL_TOKEN')) is not None:
            try:
                session.auth_with_token(token=token)
            except Exception as exc:
                raise argparse.ArgumentTypeError(f"Unabled to authenticate with the given environment's `EDL_TOKEN` (token may need to be refreshed). Original exception: {str(exc)}")
        else:
            raise argparse.ArgumentTypeError("ERROR: Environment variable `EDL_TOKEN` token not set, cannot create authenticated session for tests. Are you running this in the correct local/github action environment?")

    return session

def set_should_auth_session_with_creds(user_input: str) -> ASFSession:
    should_auth = string_to_bool(user_input)
    session = ASFSession()
    if should_auth:
        session.auth_with_creds(input('EDL Username'), getpass('EDL Password'))

    return session

def set_should_auth_session_with_token(user_input: str) -> ASFSession:
    should_auth = string_to_bool(user_input)
    session = ASFSession()
    if should_auth:
        session.auth_with_token(getpass('EDL Token'))

    return session

def string_to_bool(user_input: str) -> bool:
    user_input = str(user_input).upper()
    if 'TRUE'.startswith(user_input):
       return True
    elif 'FALSE'.startswith(user_input):
       return False
    else:
       raise argparse.ArgumentTypeError(f"ERROR: Could not convert '{user_input}' to bool (true/false/t/f).")

def pytest_addoption(parser: pytest.Parser):
    parser.addoption("--should_auth_session", action="store", dest="authenticated_session", type=set_should_auth_session, default='FALSE',
        help = "'should_auth_session': Set if the test case requires authentication (pull from `EDL_TOKEN` environment variable)"
    )

    parser.addoption("--auth_with_creds", action="store", dest="authenticated_session", type=set_should_auth_session_with_creds, default='FALSE',
                     help = "'auth_with_creds': Use EDL username and password to authenticate session for relevant tests")

    parser.addoption("--auth_with_token", action="store", dest="authenticated_session", type=set_should_auth_session_with_token, default='FALSE',
                     help = "'auth_with_creds': Use EDL token to authenticate session for relevant tests")



import copy
from typing import Dict, Union
from asf_search import ASFSearchOptions, ASFSession
from asf_search.Products import S1Product
from asf_search.CMR.translate import try_parse_date
from asf_search.CMR.translate import try_parse_int
from asf_search.constants import PRODUCT_TYPE


class S1BurstProduct(S1Product):
    """
    S1Product Subclass made specifically for Sentinel-1 SLC-BURST products

    Key features/properties:
    - `properties['burst']` contains SLC-BURST Specific fields
        such as `fullBurstID` and `burstIndex`
    - `properties['additionalUrls']` contains BURST-XML url
    - SLC-BURST specific stacking params

    ASF Dataset Documentation Page:
        https://asf.alaska.edu/datasets/data-sets/derived-data-sets/sentinel-1-bursts/
    """

    _base_properties = {
        **S1Product._base_properties,
        'bytes': {'path': ['AdditionalAttributes', ('Name', 'BYTE_LENGTH'),  'Values', 0]},
        'absoluteBurstID': {'path': ['AdditionalAttributes', ('Name', 'BURST_ID_ABSOLUTE'), 'Values', 0], 'cast': try_parse_int},
        'relativeBurstID': {'path': ['AdditionalAttributes', ('Name', 'BURST_ID_RELATIVE'), 'Values', 0], 'cast': try_parse_int},
        'fullBurstID': {'path': ['AdditionalAttributes', ('Name', 'BURST_ID_FULL'), 'Values', 0]},
        'burstIndex': {'path': ['AdditionalAttributes', ('Name', 'BURST_INDEX'), 'Values', 0], 'cast': try_parse_int},
        'samplesPerBurst': {'path': ['AdditionalAttributes', ('Name', 'SAMPLES_PER_BURST'), 'Values', 0], 'cast': try_parse_int},
        'subswath': {'path': ['AdditionalAttributes', ('Name', 'SUBSWATH_NAME'), 'Values', 0]},
        'azimuthTime': {'path': ['AdditionalAttributes', ('Name', 'AZIMUTH_TIME'), 'Values', 0], 'cast': try_parse_date},
        'azimuthAnxTime': {'path': ['AdditionalAttributes', ('Name', 'AZIMUTH_ANX_TIME'), 'Values', 0]},
    }

    def __init__(self, args: Dict = {}, session: ASFSession = ASFSession()):
        super().__init__(args, session)
        self.properties["sceneName"] = self.properties["fileID"]
        self.properties.pop('frameNumber', None)
        # Gathers burst properties into `burst` specific dict
        # rather than properties dict to limit breaking changes
        self.properties["burst"] = {
            "absoluteBurstID": self.properties.pop("absoluteBurstID"),
            "relativeBurstID": self.properties.pop("relativeBurstID"),
            "fullBurstID": self.properties.pop("fullBurstID"),
            "burstIndex": self.properties.pop("burstIndex"),
            "samplesPerBurst": self.properties.pop("samplesPerBurst"),
            "subswath": self.properties.pop("subswath"),
            "azimuthTime": self.properties.pop("azimuthTime"),
            "azimuthAnxTime": self.properties.pop("azimuthAnxTime"),
        }

        urls = self.umm_get(
            self.umm, "RelatedUrls", ("Type", [("USE SERVICE API", "URL")]), 0
        )
        if urls is not None:
            self.properties["url"] = urls[0]
            self.properties["fileName"] = (
                self.properties["fileID"] + "." + urls[0].split(".")[-1]
            )
            self.properties["additionalUrls"] = [urls[1]]  # xml-metadata url

    def get_stack_opts(self, opts: ASFSearchOptions = None):
        """
        Returns the search options asf-search will use internally
        to build an SLC-BURST baseline stack from

        :param opts: additional criteria for limiting
        :returns ASFSearchOptions used for build Sentinel-1 SLC-BURST Stack
        """
        stack_opts = ASFSearchOptions() if opts is None else copy(opts)

        stack_opts.processingLevel = self.get_default_baseline_product_type()
        stack_opts.fullBurstID = self.properties["burst"]["fullBurstID"]
        stack_opts.polarization = [self.properties["polarization"]]
        return stack_opts

    def _get_additional_filenames_and_urls(self, default_filename: str = None):
        # Burst XML filenames are just numbers, this makes it more indentifiable
        if default_filename is None:
            default_filename = self.properties["fileName"]

        file_name = f"{'.'.join(default_filename.split('.')[:-1])}.xml"

        return [(file_name, self.properties["additionalUrls"][0])]

    @staticmethod
    def get_default_baseline_product_type() -> Union[str, None]:
        """
        Returns the product type to search for when building a baseline stack.
        """
        return PRODUCT_TYPE.BURST



from typing import Dict, List, Union
from asf_search.CMR.MissionList import get_campaigns


def campaigns(platform: str) -> List[str]:
    """
    Returns a list of campaign names for the given platform,
    each name being usable as a campaign for asf_search.search() and asf_search.geo_search()

    :param platform: The name of the platform to gather campaign names for.
    Platforms currently supported include UAVSAR, AIRSAR, and SENTINEL-1 INTERFEROGRAM (BETA)

    :return: A list of campaign names for the given platform
    """
    data = {'include_facets': 'true', 'provider': 'ASF'}

    if platform is not None:
        if platform == 'UAVSAR':
            data['platform[]'] = 'G-III'
            data['instrument[]'] = 'UAVSAR'
        elif platform == 'AIRSAR':
            data['platform[]'] = 'DC-8'
            data['instrument[]'] = 'AIRSAR'
        elif platform == 'SENTINEL-1 INTERFEROGRAM (BETA)':
            data['platform[]'] = 'SENTINEL-1A'
        else:
            data['platform[]'] = platform

    missions = get_campaigns(data)
    mission_names = _get_project_names(missions)

    return mission_names


def _get_project_names(data: Union[Dict, List]) -> List[str]:
    """
    Recursively searches for campaign names
    under "Projects" key in CMR umm_json response

    :param data: CMR umm_json response

    :return: A list of found campaign names for the given platform
    """
    output = []
    if isinstance(data, Dict):
        for key, value in data.items():
            if key == 'Projects':
                return [list(item.values())[0] for item in value]
            output.extend(_get_project_names(value))
    elif isinstance(data, List):
        for item in data:
            output.extend(_get_project_names(item))

    return output




from copy import copy
from typing import Dict, Optional, Type
import warnings

from asf_search import ASFSession
from asf_search.ASFProduct import ASFProduct
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.Products import S1Product
from asf_search.CMR.translate import try_parse_float
from asf_search.constants import PRODUCT_TYPE, DATASET, POLARIZATION, BEAMMODE
from asf_search import ASFSearchResults

try:
    from asf_enumeration import aria_s1_gunw
except ImportError:
    aria_s1_gunw = None


class ARIAS1GUNWProduct(S1Product):
    """
    Used for ARIA S1 GUNW Products

    ASF Dataset Documentation Page:
        https://asf.alaska.edu/data-sets/derived-data-sets/sentinel-1-interferograms/
    """

    _base_properties = {
        **S1Product._base_properties,
        'perpendicularBaseline': {
            'path': [
                'AdditionalAttributes',
                ('Name', 'PERPENDICULAR_BASELINE'),
                'Values',
                0,
            ],
            'cast': try_parse_float,
        },
        'orbit': {'path': ['OrbitCalculatedSpatialDomains']},
        'inputGranules': {'path': ['InputGranules']},
        'ariaVersion': {'path': ['AdditionalAttributes', ('Name', 'VERSION'), 'Values', 0]},
    }

    def __init__(self, args: Dict = {}, session: ASFSession = ASFSession()):
        super().__init__(args, session)
        self.properties['orbit'] = [orbit['OrbitNumber'] for orbit in self.properties['orbit']]
        
        if self.properties.get("sceneName") is None:
            self.properties["sceneName"] = self.properties["fileID"]

        urls = self.umm_get(self.umm, 'RelatedUrls', ('Type', [('USE SERVICE API', 'URL')]), 0)

        self.properties['additionalUrls'] = []
        if urls is not None:
            self.properties['url'] = urls[0]
            self.properties['fileName'] = self.properties['fileID'] + '.' + urls[0].split('.')[-1]
            self.properties['additionalUrls'] = urls[1:]

        if self.properties.get('bytes') is None:
            bytesMap = self.umm_get(self.umm, 'DataGranule', 'ArchiveAndDistributionInformation')
            if isinstance(bytesMap, list):
                bytesMap = {entry['Format']: entry for entry in bytesMap}
                if bytesMap.get('netCDF-4') is not None:
                    self.properties['bytes'] = bytesMap['netCDF-4']['SizeInBytes']
                    self.properties['md5sum'] = bytesMap['netCDF-4']['Checksum']['Value']

    def get_stack_opts(self, opts: Optional[ASFSearchOptions] = None) -> Optional[ASFSearchOptions]:
        """
        Build search options that can be used to find an insar stack for this product

        :return: ASFSearchOptions describing appropriate options
        for building a stack from this product
        """
        return ARIAS1GUNWProduct.get_stack_opts_for_frame(self.properties['frameNumber'], opts)


    @staticmethod
    def get_stack_opts_for_frame(frame_id: int, opts: Optional[ASFSearchOptions] = None) -> Optional[ASFSearchOptions]:
        if aria_s1_gunw is None:
            warnings.warn("Failed to import asf-enumeration package. \
                          Make sure it's installed in your current environment to perform stacking with the ARIAS1GUNWProduct type")
            return None
        stack_opts = ASFSearchOptions() if opts is None else copy(opts)
        aria_frame = aria_s1_gunw.get_frame(frame_id)
        
        # pulled from asf-enumeration package implementation
        stack_opts.dataset = DATASET.SENTINEL1
        stack_opts.platform = ['SA', 'SB', 'SC']
        stack_opts.processingLevel = PRODUCT_TYPE.SLC
        stack_opts.beamMode = BEAMMODE.IW
        stack_opts.polarization = [POLARIZATION.VV, POLARIZATION.VV_VH]
        stack_opts.flightDirection = aria_frame.flight_direction
        stack_opts.relativeOrbit = aria_frame.path
        stack_opts.intersectsWith = aria_frame.wkt

        return stack_opts

    def is_valid_reference(self):
        return False

    @staticmethod
    def get_default_baseline_product_type() -> str:
        """
        Returns the product type to search for when building a baseline stack.
        """
        return PRODUCT_TYPE.SLC

    def stack(
        self, opts: Optional[ASFSearchOptions] = None, useSubclass: Type['ASFProduct'] = None
    ) -> ASFSearchResults:
        from asf_search.baseline import get_baseline_from_stack
        s1_products = self.get_aria_groups_for_frame(self.properties['frameNumber'])
        
        if len(s1_products) == 0:
            reference = None
        else:
            reference = s1_products[0]

        target_stack, warnings = get_baseline_from_stack(reference, s1_products)

        return target_stack

    @staticmethod
    def _is_subclass(item: Dict) -> bool:
        platform = ASFProduct.umm_get(item['umm'], 'Platforms', 0, 'ShortName')
        if platform in ['SENTINEL-1A', 'SENTINEL-1B', 'SENTINEL-1C']:
            asf_platform = ASFProduct.umm_get(
                item['umm'],
                'AdditionalAttributes',
                ('Name', 'ASF_PLATFORM'),
                'Values',
                0,
            )
            return 'Sentinel-1 Interferogram' in asf_platform

        return False

    @staticmethod
    def get_aria_groups_for_frame(frame: str) -> ASFSearchResults:
        """Returns the sentinel1 acquisitions over a given frame that overlap the aria frame with the given aria frame ID,
        filter products from groups that overlap the frame by < 90%"""
        if aria_s1_gunw is None:
            raise ImportError(
            'Could not find asf-enumeration package in current python environment. '
            '"asf-enumeration" is an optional dependency of asf-search required '
            'for stacking with the ARIAS1GUNWProduct type. '
            'Enable by including the appropriate pip or conda install. '
            'Ex: `python3 -m pip install asf-search[asf-enumeration]`'
        )
        aria_frame = aria_s1_gunw.get_frame(frame_id=int(frame))
        groups = aria_s1_gunw.get_acquisitions(aria_frame)
        output = []

        # ignore acquisition groups with 0 products
        # take the first product that has a baseline value, otherwise just take the first product in the group
        for group in groups:
            if len(group.products) == 0:
                continue
            valid = next((product for product in group.products if product.has_baseline()), group.products[0])
            output.append(valid)

        return ASFSearchResults(output)



from typing import Dict, Union
from asf_search import ASFSession, ASFStackableProduct
from asf_search.constants import PRODUCT_TYPE


class JERSProduct(ASFStackableProduct):
    """
    ASF Dataset Documentation Page: https://asf.alaska.edu/datasets/daac/jers-1/
    """

    _base_properties = {
        **ASFStackableProduct._base_properties,
        'browse': {'path': ['RelatedUrls', ('Type', [('GET RELATED VISUALIZATION', 'URL')])]},
        'groupID': {'path': ['AdditionalAttributes', ('Name', 'GROUP_ID'), 'Values', 0]},
        'md5sum': {'path': ['AdditionalAttributes', ('Name', 'MD5SUM'), 'Values', 0]},
        'beamModeType': {'path': ['AdditionalAttributes', ('Name', 'BEAM_MODE_TYPE'), 'Values', 0]},
        'insarStackId': {'path': ['AdditionalAttributes', ('Name', 'INSAR_STACK_ID'), 'Values', 0]},
    }

    def __init__(self, args: Dict = {}, session: ASFSession = ASFSession()):
        super().__init__(args, session)

    @staticmethod
    def get_default_baseline_product_type() -> Union[str, None]:
        """
        Returns the product type to search for when building a baseline stack.
        """
        return PRODUCT_TYPE.L0



class ASFError(Exception):
    """Base ASF Exception, not intended for direct use"""


class ASFSearchError(ASFError):
    """Base search-related Exception"""


class ASFSearch4xxError(ASFSearchError):
    """Raise when CMR returns a 4xx error"""


class ASFSearch5xxError(ASFSearchError):
    """Raise when CMR returns a 5xx error"""


class ASFBaselineError(ASFSearchError):
    """Raise when baseline related errors occur"""


class ASFDownloadError(ASFError):
    """Base download-related Exception"""


class ASFAuthenticationError(ASFError):
    """Base download-related Exception"""


class ASFWKTError(ASFError):
    """Raise when wkt related errors occur"""


class CoherenceEstimationError(ASFError):
    """Raise if coherence estimation is requested for a Pair with a temporal baseline > 48 days"""


class CMRError(Exception):
    """Base CMR Exception"""


class CMRConceptIDError(CMRError):
    """Raise when CMR encounters a concept-id error"""


class CMRIncompleteError(CMRError):
    """Raise when CMR returns an incomplete page of results"""



import inspect
import json
from types import GeneratorType

from asf_search import ASF_LOGGER
from .jsonlite import JSONLiteStreamArray

def results_to_jsonlite2(results):
    ASF_LOGGER.info('started translating results to jsonlite2 format')
    
    if len(results) == 0:
            yield from json.JSONEncoder(indent=2, sort_keys=True).iterencode({'results': []})
            return
        
    if not inspect.isgeneratorfunction(results) and not isinstance(results, GeneratorType):
        results = [results]

    streamer = JSONLite2StreamArray(results)

    for p in json.JSONEncoder(sort_keys=True, separators=(",", ":")).iterencode(
        {"results": streamer}
    ):
        yield p


class JSONLite2StreamArray(JSONLiteStreamArray):
    def getItem(self, p):
        # pre-processing of the result is the same as in the base jsonlite streamer,
        # so use that and then rename/substitute fields
        p = super().getItem(p)
        result = {
            "b": [a.replace(p.get("granuleName"), "{gn}") for a in p.get("browse")]
            if p["browse"] is not None
            else p["browse"],
            "bm": p.get("beamMode"),
            "d": p.get("dataset"),
            "du": p.get("downloadUrl").replace(p.get("granuleName"), "{gn}"),
            "f": p.get("frame"),
            "fd": p.get("flightDirection"),
            "fl": p.get("flightLine"),
            "fn": p.get("fileName").replace(p.get("granuleName"), "{gn}"),
            "fr": p.get("faradayRotation"),  # ALOS
            "gid": p.get("groupID").replace(p.get("granuleName"), "{gn}"),
            "gn": p.get("granuleName"),
            "i": p.get("instrument"),
            "in": p.get("canInSAR"),
            "mn": p.get("missionName"),
            "o": p.get("orbit"),
            "on": p.get("offNadirAngle"),  # ALOS
            "p": p.get("path"),
            "pid": p.get("productID").replace(p.get("granuleName"), "{gn}"),
            "pa": p.get("pointingAngle"),
            "po": p.get("polarization"),
            "pt": p.get("productType"),
            "ptd": p.get("productTypeDisplay"),
            "s": p.get("sizeMB"),
            "ss": p.get("stackSize"),  # Used for datasets with precalculated stacks
            "st": p.get("startTime"),
            "stp": p.get("stopTime"),
            "t": p.get("thumb").replace(p.get("granuleName"), "{gn}")
            if p.get("thumb") is not None
            else p.get("thumb"),
            "w": p.get("wkt"),
            "wu": p.get("wkt_unwrapped"),
            "pge": p.get("pgeVersion"),
            "adu": p.get("additionalUrls"),
            's3u': p.get("s3Urls"),
        }

        if 'temporalBaseline' in p.keys():
            result['tb'] = p['temporalBaseline']
        if 'perpendicularBaseline' in p.keys():
            result['pb'] = p['perpendicularBaseline']

        if p.get('burst') is not None: # is a burst product
            result['s1b'] = p['burst']
            result['f'] = None

        if p.get('opera') is not None:
            result['s1o'] = p['opera']
        
        if p.get('nisar') is not None:
            result['nsr'] = p['nisar']
            result["cnm"] = p["collectionName"]
            result["cid"] = p["conceptID"]
        
        if p.get('ariaVersion') is not None:
            result['ariav'] = p.get('ariaVersion')
    
        return result

    def getOutputType(self) -> str:
        return "jsonlite2"



import inspect
import json
from types import GeneratorType

from asf_search import ASF_LOGGER


def results_to_geojson(results):
    ASF_LOGGER.info('started translating results to geojson format')

    if not inspect.isgeneratorfunction(results) and not isinstance(results, GeneratorType):
        results = [results]

    streamer = GeoJSONStreamArray(results)

    for p in json.JSONEncoder(indent=2, sort_keys=True).iterencode(
        {'type': 'FeatureCollection', 'features': streamer}
    ):
        yield p


class GeoJSONStreamArray(list):
    def __init__(self, results):
        self.results = results

        # need to make sure we actually have results so we can intelligently set __len__, otherwise
        # iterencode behaves strangely and will output invalid json
        self.len = 1

    def __iter__(self):
        return self.streamDicts()

    def __len__(self):
        return self.len

    def streamDicts(self):
        completed = False
        for page_idx, page in enumerate(self.results):
            ASF_LOGGER.info(f'Streaming {len(page)} products from page {page_idx}')
            completed = page.searchComplete

            yield from [self.getItem(p) for p in page if p is not None]

        if not completed:
            ASF_LOGGER.warn('Failed to download all results from CMR')

        ASF_LOGGER.info('Finished streaming geojson results')

    def getItem(self, p):
        return p.geojson()



import csv
import os
from types import GeneratorType
from urllib import parse
from asf_search import ASF_LOGGER

from asf_search.export.export_translators import ASFSearchResults_to_properties_list
import inspect

extra_csv_fields = [
    ('sceneDate', ['AdditionalAttributes', ('Name', 'ACQUISITION_DATE'), 'Values', 0]),
    ('nearStartLat', ['AdditionalAttributes', ('Name', 'NEAR_START_LAT'), 'Values', 0]),
    ('nearStartLon', ['AdditionalAttributes', ('Name', 'NEAR_START_LON'), 'Values', 0]),
    ('farStartLat', ['AdditionalAttributes', ('Name', 'FAR_START_LAT'), 'Values', 0]),
    ('farStartLon', ['AdditionalAttributes', ('Name', 'FAR_START_LON'), 'Values', 0]),
    ('nearEndLat', ['AdditionalAttributes', ('Name', 'NEAR_END_LAT'), 'Values', 0]),
    ('nearEndLon', ['AdditionalAttributes', ('Name', 'NEAR_END_LON'), 'Values', 0]),
    ('farEndLat', ['AdditionalAttributes', ('Name', 'FAR_END_LAT'), 'Values', 0]),
    ('farEndLon', ['AdditionalAttributes', ('Name', 'FAR_END_LON'), 'Values', 0]),
    (
        'faradayRotation',
        ['AdditionalAttributes', ('Name', 'FARADAY_ROTATION'), 'Values', 0],
    ),
    (
        'configurationName',
        ['AdditionalAttributes', ('Name', 'BEAM_MODE_DESC'), 'Values', 0],
    ),
    ('doppler', ['AdditionalAttributes', ('Name', 'DOPPLER'), 'Values', 0]),
    ('sizeMB', ['DataGranule', 'ArchiveAndDistributionInformation', 0, 'Size']),
    (
        'insarStackSize',
        ['AdditionalAttributes', ('Name', 'INSAR_STACK_SIZE'), 'Values', 0],
    ),
    (
        'offNadirAngle',
        ['AdditionalAttributes', ('Name', 'OFF_NADIR_ANGLE'), 'Values', 0],
    ),
]

fieldnames = (
    'Granule Name',
    'Platform',
    'Sensor',
    'Beam Mode',
    'Beam Mode Description',
    'Orbit',
    'Path Number',
    'Frame Number',
    'Acquisition Date',
    'Processing Date',
    'Processing Level',
    'Start Time',
    'End Time',
    'Center Lat',
    'Center Lon',
    'Near Start Lat',
    'Near Start Lon',
    'Far Start Lat',
    'Far Start Lon',
    'Near End Lat',
    'Near End Lon',
    'Far End Lat',
    'Far End Lon',
    'Faraday Rotation',
    'Ascending or Descending?',
    'URL',
    'Size (MB)',
    'Off Nadir Angle',
    'Stack Size',
    'Doppler',
    'GroupID',
    'Pointing Angle',
    'TemporalBaseline',
    'PerpendicularBaseline',
    'relativeBurstID',
    'absoluteBurstID',
    'fullBurstID',
    'burstIndex',
    'azimuthTime',
    'azimuthAnxTime',
    'samplesPerBurst',
    'subswath',
)


def results_to_csv(results):
    ASF_LOGGER.info('started translating results to csv format')

    if inspect.isgeneratorfunction(results) or isinstance(results, GeneratorType):
        return CSVStreamArray(results)

    return CSVStreamArray([results])


class CSVStreamArray(list):
    def __init__(self, results):
        self.pages = results
        self.len = 1

    def __iter__(self):
        return self.streamRows()

    def __len__(self):
        return self.len

    def get_additional_output_fields(self, product):
        additional_fields = {}
        for key, path in extra_csv_fields:
            additional_fields[key] = product.umm_get(product.umm, *path)

        return additional_fields

    def streamRows(self):
        f = CSVBuffer()
        writer = csv.DictWriter(f, quoting=csv.QUOTE_ALL, fieldnames=fieldnames)
        yield writer.writeheader()

        completed = False
        for page_idx, page in enumerate(self.pages):
            ASF_LOGGER.info(f'Streaming {len(page)} products from page {page_idx}')
            completed = page.searchComplete

            properties_list = ASFSearchResults_to_properties_list(
                page, self.get_additional_output_fields
            )
            yield from [writer.writerow(self.getItem(p)) for p in properties_list]

        if not completed:
            ASF_LOGGER.warn('Failed to download all results from CMR')

        ASF_LOGGER.info('Finished streaming csv results')

    def getItem(self, p):
        if p.get('sizeMB') is None and p.get('platform') == 'NISAR':
            if isinstance(p.get('bytes'), dict):
                a = parse.urlparse(p['url'])
                file_name = os.path.basename(a.path)
                bytes_entry = p['bytes'].get(file_name)
                if bytes_entry is not None:          
                    size_mb = bytes_entry['bytes'] / 1000000 
                p['sizeMB'] = str(size_mb) if size_mb < 0.01 else "{:10.2f}".format(size_mb)

        return {
            'Granule Name': p.get('sceneName'),
            'Platform': p.get('platform'),
            'Sensor': p.get('sensor'),
            'Beam Mode': p.get('beamModeType'),
            'Beam Mode Description': p.get('configurationName'),
            'Orbit': p.get('orbit'),
            'Path Number': p.get('pathNumber'),
            'Frame Number': p.get('frameNumber'),
            'Acquisition Date': p.get('sceneDate'),
            'Processing Date': p.get('processingDate'),
            'Processing Level': p.get('processingLevel'),
            'Start Time': p.get('startTime'),
            'End Time': p.get('stopTime'),
            'Center Lat': p.get('centerLat'),
            'Center Lon': p.get('centerLon'),
            'Near Start Lat': p.get('nearStartLat'),
            'Near Start Lon': p.get('nearStartLon'),
            'Far Start Lat': p.get('farStartLat'),
            'Far Start Lon': p.get('farStartLon'),
            'Near End Lat': p.get('nearEndLat'),
            'Near End Lon': p.get('nearEndLon'),
            'Far End Lat': p.get('farEndLat'),
            'Far End Lon': p.get('farEndLon'),
            'Faraday Rotation': p.get('faradayRotation'),
            'Ascending or Descending?': p.get('flightDirection'),
            'URL': p.get('url'),
            'Size (MB)': p.get('sizeMB'),
            'Off Nadir Angle': p.get('offNadirAngle'),
            'Stack Size': p.get('insarStackSize'),
            'Doppler': p.get('doppler'),
            'GroupID': p.get('groupID'),
            'Pointing Angle': p.get('pointingAngle'),
            'TemporalBaseline': p.get('teporalBaseline'),
            'PerpendicularBaseline': p.get('pependicularBaseline'),
            'relativeBurstID': p.get('burst', {}).get('relativeBurstID')
            if p.get('processingLevel') == 'BURST' else None,
            'absoluteBurstID': p['burst']['absoluteBurstID']
            if p.get('processingLevel') == 'BURST'
            else None,
            'fullBurstID': p['burst']['fullBurstID']
            if p.get('processingLevel') == 'BURST'
            else None,
            'burstIndex': p['burst']['burstIndex'] if p.get('processingLevel') == 'BURST' else None,
            'azimuthTime': p['burst']['azimuthTime']
            if p.get('processingLevel') == 'BURST'
            else None,
            'azimuthAnxTime': p['burst']['azimuthAnxTime']
            if p.get('processingLevel') == 'BURST'
            else None,
            'samplesPerBurst': p['burst']['samplesPerBurst']
            if p.get('processingLevel') == 'BURST'
            else None,
            'subswath': p['burst']['subswath'] if p.get('processingLevel') == 'BURST' else None,
        }


class CSVBuffer:
    # https://docs.djangoproject.com/en/3.2/howto/outputting-csv/#streaming-large-csv-files
    # A dummy CSV buffer to be used by the csv.writer class, returns the
    # formatted csv row "written" to it when writer.writerow/writeheader is called

    def write(self, value):
        """Write the value by returning it, instead of storing in a buffer."""
        return value



"""asf_search setuptools configuration"""

from setuptools import find_packages, setup

requirements = [
    'requests',
    'shapely',
    'pytz',
    'numpy',
    'dateparser',
    'python-dateutil',
    'tenacity>=8.2.2',
]

test_requirements = [
    'pytest==8.1.1',
    'pytest-automation==3.0.0',
    'pytest-cov',
    'pytest-xdist',
    'coverage',
    'requests-mock==1.11.0',
    'nbformat',
    'nbconvert',
    'ipykernel',
]

extra_requirements = [
    'remotezip>=0.10.0',
    'ciso8601',
]

# Required for ARIA-S1 GUNW Stacking
asf_enumeration = [
    'asf-enumeration>=0.4.0'
]

# Required for optional Sentinel-1 Pair coherence estimation
coherence = [
    'pandas',
    'zarr',
    's3fs',
    'rioxarray',
]

with open('README.md', 'r') as readme_file:
    readme = readme_file.read()

setup(
    name='asf_search',
    # version=Declared in pyproject.toml, through "[tool.setuptools_scm]"
    author='Alaska Satellite Facility Discovery Team',
    author_email='uaf-asf-discovery@alaska.edu',
    description="Python wrapper for ASF's SearchAPI",
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/asfadmin/Discovery-asf_search.git',
    project_urls={'Documentation': 'https://docs.asf.alaska.edu/asf_search/basics/'},
    packages=find_packages(exclude=['tests.*', 'tests', 'examples.*', 'examples']),
    package_dir={'asf_search': 'asf_search'},
    include_package_data=True,
    python_requires='>=3.10',
    install_requires=requirements,
    extras_require={'test': test_requirements, 'extras': extra_requirements, 'asf-enumeration': asf_enumeration, 'coherence': coherence},
    license='BSD',
    license_files=('LICENSE',),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering :: Atmospheric Science',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Scientific/Engineering :: Hydrology',
        'Topic :: Utilities',
    ],
)



from enum import Enum
import copy
from typing import Dict, Union
from asf_search import ASFSession, ASFProduct
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.exceptions import ASFBaselineError


class ASFStackableProduct(ASFProduct):
    """
    Used for ERS-1 and ERS-2 products

    ASF ERS-1 Dataset Documentation Page: https://asf.alaska.edu/datasets/daac/ers-1/
    ASF ERS-2 Dataset Documentation Page: https://asf.alaska.edu/datasets/daac/ers-2/
    """

    class BaselineCalcType(Enum):
        """
        Defines how asf-search will calculate perpendicular baseline for products of this subclass
        """

        PRE_CALCULATED = 0
        """Has pre-calculated insarBaseline value that will be used for perpendicular calculations"""  # noqa F401
        CALCULATED = 1
        """Uses position/velocity state vectors and ascending node time for perpendicular calculations"""  # noqa F401

    baseline_type = BaselineCalcType.PRE_CALCULATED
    """Determines how asf-search will attempt to stack products of this type."""

    def __init__(self, args: Dict = {}, session: ASFSession = ASFSession()):
        super().__init__(args, session)
        self.baseline = self.get_baseline_calc_properties()

    def get_baseline_calc_properties(self) -> Dict:
        insarBaseline = self.umm_cast(
            float,
            self.umm_get(
                self.umm,
                'AdditionalAttributes',
                ('Name', 'INSAR_BASELINE'),
                'Values',
                0,
            ),
        )

        if insarBaseline is None:
            return None

        return {'insarBaseline': insarBaseline}

    def get_stack_opts(self, opts: ASFSearchOptions = None):
        stack_opts = ASFSearchOptions() if opts is None else copy(opts)
        stack_opts.processingLevel = self.get_default_baseline_product_type()

        if self.properties.get('insarStackId') in [None, 'NA', 0, '0']:
            raise ASFBaselineError(
                'Requested reference product needs a baseline stack ID '
                f'but does not have one: {self.properties["fileID"]}'
            )

        stack_opts.insarStackId = self.properties['insarStackId']
        return stack_opts

    def is_valid_reference(self):
        # we don't stack at all if any of stack is missing insarBaseline,
        # unlike stacking S1 products(?)
        if 'insarBaseline' not in self.baseline:
            raise ValueError('No baseline values available for precalculated dataset')

        return True

    @staticmethod
    def get_default_baseline_product_type() -> Union[str, None]:
        """
        Returns the product type to search for when building a baseline stack.
        """
        return None

    def has_baseline(self) -> bool:
        baseline = self.get_baseline_calc_properties()

        return baseline is not None



import inspect
import json
import re
from types import GeneratorType
from typing import Tuple
from shapely.geometry import shape
from shapely.ops import transform

from asf_search import ASF_LOGGER
from asf_search.export.export_translators import ASFSearchResults_to_properties_list

_MB = 1048576

extra_jsonlite_fields = [
    (
        "processingTypeDisplay",
        ["AdditionalAttributes", ("Name", "PROCESSING_TYPE_DISPLAY"), "Values", 0],
    ),
    ("thumb", ["AdditionalAttributes", ("Name", "THUMBNAIL_URL"), "Values", 0]),
    (
        "faradayRotation",
        ["AdditionalAttributes", ("Name", "FARADAY_ROTATION"), "Values", 0],
    ),
    ("sizeMB", ["DataGranule", "ArchiveAndDistributionInformation", 0, "Size"]),
    ("flightLine", ["AdditionalAttributes", ("Name", "FLIGHT_LINE"), "Values", 0]),
    ("missionName", ["AdditionalAttributes", ("Name", "MISSION_NAME"), "Values", 0]),
]

def results_to_jsonlite(results):
    ASF_LOGGER.info('started translating results to jsonlite format')
    if len(results) == 0:
        yield from json.JSONEncoder(indent=2, sort_keys=True).iterencode({'results': []})
        return

    if not inspect.isgeneratorfunction(results) and not isinstance(results, GeneratorType):
        results = [results]

    streamer = JSONLiteStreamArray(results)
    jsondata = {"results": streamer}

    for p in json.JSONEncoder(indent=2, sort_keys=True).iterencode(jsondata):
        yield p


def unwrap_shape(x, y, z=None):
    x = x if x > 0 else x + 360
    return tuple([x, y])


def get_wkts(geometry) -> Tuple[str, str]:
    wrapped = shape(geometry)

    min_lon, max_lon = (wrapped.bounds[0], wrapped.bounds[2])

    if max_lon - min_lon > 180:
        unwrapped = transform(unwrap_shape, wrapped)
    else:
        unwrapped = wrapped

    return wrapped.wkt, unwrapped.wkt


class JSONLiteStreamArray(list):
    def __init__(self, results):
        self.results = results

        # need to make sure we actually have results so we can intelligently set __len__, otherwise
        # iterencode behaves strangely and will output invalid json
        self.len = 1

    def __iter__(self):
        return self.streamDicts()

    def __len__(self):
        return self.len

    def get_additional_output_fields(self, product):
        # umm = product.umm

        additional_fields = {}
        for key, path in extra_jsonlite_fields:
            additional_fields[key] = product.umm_get(product.umm, *path)

        platform = product.properties.get("platform")
        if platform is None:
            platform = ""
        if platform.upper() in [
            "ALOS",
            "RADARSAT-1",
            "JERS-1",
            "ERS-1",
            "ERS-2",
        ]:
            insarGrouping = product.umm_get(
                product.umm,
                *["AdditionalAttributes", ("Name", "INSAR_STACK_ID"), "Values", 0],
            )

            if insarGrouping not in [None, 0, "0", "NA", "NULL"]:
                additional_fields["canInsar"] = True
                additional_fields["insarStackSize"] = product.umm_get(
                    product.umm,
                    *[
                        "AdditionalAttributes",
                        ("Name", "INSAR_STACK_SIZE"),
                        "Values",
                        0,
                    ],
                )
            else:
                additional_fields["canInsar"] = False
        else:
            additional_fields["canInsar"] = product.baseline is not None

        additional_fields["geometry"] = product.geometry

        return additional_fields

    def streamDicts(self):
        completed = False
        for page_idx, page in enumerate(self.results):
            ASF_LOGGER.info(f"Streaming {len(page)} products from page {page_idx}")
            completed = page.searchComplete

            yield from [
                self.getItem(p)
                for p in ASFSearchResults_to_properties_list(
                    page, self.get_additional_output_fields
                )
                if p is not None
            ]

        if not completed:
            ASF_LOGGER.warn("Failed to download all results from CMR")

        ASF_LOGGER.info(f"Finished streaming {self.getOutputType()} results")

    def getItem(self, p):
        for i in p.keys():
            if p[i] == "NA" or p[i] == "":
                p[i] = None
        try:
            if p.get("offNadirAngle") is not None and float(p["offNadirAngle"]) < 0:
                p["offNadirAngle"] = None
        except TypeError:
            pass

        try:
            if p.get("patNumber"):
                if float(p["pathNumber"]) < 0:
                    p["pathNumber"] = None
        except TypeError:
            pass

        try:
            if p.get("groupID") is None:
                p["groupID"] = p["sceneName"]
        except TypeError:
            pass

        try:
            p["sizeMB"] = float(p["sizeMB"])
        except TypeError:
            pass

        try:
            p["pathNumber"] = int(p["pathNumber"])
        except TypeError:
            pass

        try:
            p['frameNumber'] = int(p.get('frameNumber'))
        except TypeError:
            pass

        try:
            p["orbit"] = int(p["orbit"])
        except TypeError:
            pass

        wrapped, unwrapped = get_wkts(p["geometry"])
        result = {
            "beamMode": p["beamModeType"],
            "browse": [] if p.get("browse") is None else p.get("browse"),
            "canInSAR": p.get("canInsar"),
            "dataset": p.get("platform"),
            "downloadUrl": p.get("url"),
            "faradayRotation": p.get("faradayRotation"),  # ALOS
            "fileName": p.get("fileName"),
            "flightDirection": p.get("flightDirection"),
            "flightLine": p.get("flightLine"),
            "frame": p.get("frameNumber"),
            "granuleName": p.get("sceneName"),
            "groupID": p.get("groupID"),
            "instrument": p.get("sensor"),
            "missionName": p.get("missionName"),
            "offNadirAngle": str(p["offNadirAngle"])
            if p.get("offNadirAngle") is not None
            else None,  # ALOS
            "orbit": p.get("orbit") if isinstance(p.get("orbit"), list) else [str(p["orbit"])],
            "path": p.get("pathNumber"),
            "polarization": p.get("polarization"),
            "pointingAngle": p.get("pointingAngle"),
            "productID": p.get("fileID"),
            "productType": p.get("processingLevel"),
            "productTypeDisplay": p.get("processingTypeDisplay"),
            "sizeMB": p.get("sizeMB"),
            "stackSize": p.get(
                "insarStackSize"
            ),  # Used for datasets with precalculated stacks
            "startTime": p.get("startTime"),
            "stopTime": p.get("stopTime"),
            "thumb": p.get("thumb"),
            "wkt": wrapped,
            "wkt_unwrapped": unwrapped,
            "pgeVersion": p.get("pgeVersion"),
        }

        for key in result.keys():
            if result[key] in ["NA", "NULL"]:
                result[key] = None

        if "temporalBaseline" in p.keys():
            result["temporalBaseline"] = p["temporalBaseline"]
        if "perpendicularBaseline" in p.keys():
            result["perpendicularBaseline"] = p["perpendicularBaseline"]

        if p.get("processingLevel") == "BURST":  # is a burst product
            result["burst"] = p["burst"]
            result["sizeMB"] = float(p["bytes"]) / 1024000

        elif p.get('operaBurstID') is not None or result['productID'].startswith('OPERA'):
            result['opera'] = {
                'operaBurstID': p.get('operaBurstID'),
                's3Urls': p.get('s3Urls', []),
                'additionalUrls': p.get('additionalUrls'),
            }
            if p.get('validityStartDate'):
                result['opera']['validityStartDate'] = p.get('validityStartDate')
        elif p.get('platform') == 'NISAR':
            result['nisar'] = {
                'additionalUrls': p.get('additionalUrls', []),
                's3Urls': p.get('s3Urls', []),
                'pgeVersion':  p.get('pgeVersion'),
                'crid': p.get('crid'),
                'mainBandPolarization':  p.get('mainBandPolarization'),
                'sideBandPolarization':  p.get('sideBandPolarization'),
                'frameCoverage':  p.get('frameCoverage'),
                'jointObservation':  p.get('jointObservation'),
                'rangeBandwidth':  p.get('rangeBandwidth'),
                'sizeMB': p.get('bytes'),
            }
            result["collectionName"] = p.get("collectionName")
            result["conceptID"] = p.get("conceptID")
        elif p.get('platform') == 'SEASAT 1':
            result['additionalUrls'] = p.get('additionalUrls', [])
            result['s3Urls'] = p.get('s3Urls', [])
            result['sizeMB'] = p.get('bytes', {})

        elif result.get('productID', result.get('fileName', '')).startswith('S1-GUNW'):
            result.pop("perpendicularBaseline", None)
            if p.get('ariaVersion') is None:
                version_unformatted = result.get('productID').split('v')[-1]
                result['ariaVersion'] = re.sub(r'[^0-9\.]', '', version_unformatted.replace("_", '.'))
            else:
                result['ariaVersion'] = p.get('ariaVersion')
                result['productTypeDisplay'] = 'Standard Product, NetCDF'
            
            if result['sizeMB'] is None:
                result["sizeMB"] = float(p["bytes"]) / _MB
                pass
        
        return result

    def getOutputType(self) -> str:
        return "jsonlite"
